import pytest
import torch
import torchelie.callbacks as tcb
from torch.utils.data import DataLoader
from torchvision.datasets import FashionMNIST
from torchvision.transforms import PILToTensor


@pytest.mark.require_tensorboard
def test_tesorboard():
    from torchelie.recipes import Recipe

    batch_size = 4

    class Dataset:
        def __init__(self, batch_size):
            self.batch_size = batch_size
            self.mnist = FashionMNIST('.', download=True, transform=PILToTensor())
            self.classes = self.mnist.classes
            self.num_classes = len(self.mnist.class_to_idx)
            self.target_by_classes = [[idx for idx in range(len(self.mnist)) if self.mnist.targets[idx] == i]
                                      for i in range(self.num_classes)]

        def __len__(self):
            return self.batch_size * self.num_classes

        def __getitem__(self, item):
            idx = self.target_by_classes[item//self.batch_size][item]
            x, y = self.mnist[idx]
            x = torch.stack(3*[x]).squeeze()
            x[2] = 0
            return x, y

    dst = Dataset(batch_size)

    def train(b):
        x, y = b
        return {'letter_number_int':     int(y[0]),
                'letter_number_tensor':  y[0],
                'letter_text':  dst.classes[int(y[0])],
                'test_html':  '<b>test HTML</b>',
                'letter_gray_img_HW':   x[0, 0],
                'letter_gray_img_CHW':   x[0, :1],
                'letter_gray_imgs_NCHW':  x[:, :1],
                'letter_color_img_CHW':  x[0],
                'letter_color_imgs_NCHW': x}

    r = Recipe(train, DataLoader(dst, batch_size))
    r.callbacks.add_callbacks([
        tcb.Counter(),
        tcb.TensorboardLogger(log_every=1),
        tcb.Log('letter_number_int', 'letter_number_int'),
        tcb.Log('letter_number_tensor', 'letter_number_tensor'),
        tcb.Log('letter_text', 'letter_text'),
        tcb.Log('test_html', 'test_html'),
        tcb.Log('letter_gray_img_HW', 'letter_gray_img_HW'),
        tcb.Log('letter_gray_img_CHW', 'letter_gray_img_CHW'),
        tcb.Log('letter_gray_imgs_NCHW', 'letter_gray_imgs_NCHW'),
        tcb.Log('letter_color_img_CHW', 'letter_color_img_CHW'),
        tcb.Log('letter_color_imgs_NCHW', 'letter_color_imgs_NCHW'),
    ])
    r.run(1)


if __name__ == '__main__':
    test_tesorboard()



from typing import Dict, List, Union

import numpy as np
import torch
from scvi.train import TrainingPlan
from torch.optim.lr_scheduler import CosineAnnealingLR, StepLR

from ._constants import LOSS_KEYS
from ._module import HiClipModule
from ._utils import hiclip_metric


class HiClipTrainingPlan(TrainingPlan):
    """Training plan for the hiclip model."""

    def __init__(
        self,
        module: HiClipModule,
        n_epochs_warmup: Union[int, None] = None,
        checkpoint_freq: int = 20,
        lr=1e-4,
        weight_decay=1e-4,
        step_size_lr: int = 4,
        cosine_scheduler: bool = False,
        scheduler_max_epochs: int = 1000,
        scheduler_final_lr: float = 1e-5,
    ):
        super().__init__(
            module=module,
            lr=lr,
            weight_decay=weight_decay,
            n_epochs_kl_warmup=None,
            reduce_lr_on_plateau=False,
            lr_factor=None,
            lr_patience=None,
            lr_threshold=None,
            lr_scheduler_metric=None,
            lr_min=None,
        )

        self.n_epochs_warmup = n_epochs_warmup if n_epochs_warmup is not None else 0

        self.checkpoint_freq = checkpoint_freq

        self.scheduler = CosineAnnealingLR if cosine_scheduler else StepLR
        self.scheduler_params = (
            {"T_max": scheduler_max_epochs, "eta_min": scheduler_final_lr}
            if cosine_scheduler
            else {"step_size": step_size_lr}
        )

        self.step_size_lr = step_size_lr

        self.automatic_optimization = False
        self.iter_count = 0
        self.training_step_outputs = []
        self.validation_step_outputs = []
        self._epoch_keys = []

        self.epoch_keys = [
            "hiclip_metric",
            LOSS_KEYS.L1,
            LOSS_KEYS.PERCEPTUAL,
        ]

        self.epoch_history = {"mode": [], "epoch": []}
        for key in self.epoch_keys:
            self.epoch_history[key] = []

    def configure_optimizers(self):
        """Set up optimizers."""
        optimizers = []
        schedulers = []

        optimizers.append(
            torch.optim.Adam(
                [
                    {
                        "params": list(
                            filter(
                                lambda p: p.requires_grad,
                                self.module.parameters(),
                            )
                        ),
                        "lr": self.lr,
                        "weight_decay": self.weight_decay,
                        # betas=(0.5, 0.999),
                    }
                ]
            )
        )

        if self.step_size_lr is not None:
            for optimizer in optimizers:
                schedulers.append(self.scheduler(optimizer, **self.scheduler_params))
            return optimizers, schedulers
        else:
            return optimizers

    @property
    def epoch_keys(self):
        """Epoch keys getter."""
        return self._epoch_keys

    @epoch_keys.setter
    def epoch_keys(self, epoch_keys: List):
        self._epoch_keys.extend(epoch_keys)

    def training_step(self, batch):
        """Training step."""
        optimizers = self.optimizers()
        if not isinstance(optimizers, list):
            optimizers = [optimizers]
        # model update
        for optimizer in optimizers:
            optimizer.zero_grad()

        _, losses = self.module.forward(batch)

        self.manual_backward(losses[LOSS_KEYS.L1])
        self.manual_backward(losses[LOSS_KEYS.PERCEPTUAL])
        for optimizer in optimizers:
            optimizer.step()

        results = {
            LOSS_KEYS.L1: losses[LOSS_KEYS.L1].item(),
            LOSS_KEYS.PERCEPTUAL: losses[LOSS_KEYS.PERCEPTUAL].item(),
        }

        self.iter_count += 1

        for key in self.epoch_keys:
            if key not in results:
                results.update({key: 0.0})

        self.training_step_outputs.append(results)
        return results

    def on_train_epoch_end(self):
        """Training epoch end."""
        outputs = self.training_step_outputs
        self.epoch_history["epoch"].append(self.current_epoch)
        self.epoch_history["mode"].append("train")

        for key in self.epoch_keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))
            self.log(key, self.epoch_history[key][-1], prog_bar=True)

        if self.current_epoch > 1 and self.current_epoch % self.step_size_lr == 0:
            schedulers = self.lr_schedulers()
            if not isinstance(schedulers, list):
                schedulers = [schedulers]
            for scheduler in schedulers:
                scheduler.step()

        self.training_step_outputs.clear()

    def validation_step(self, batch, batch_idx):
        """Validation step."""
        _, losses = self(batch)

        results = {}
        for key in losses:
            results.update({key: losses[key].item()})

        results.update(
            {
                "hiclip_metric": hiclip_metric(
                    results[LOSS_KEYS.L1], results[LOSS_KEYS.PERCEPTUAL]
                )
            }
        )

        self.validation_step_outputs.append(results)
        return results

    def on_validation_epoch_end(self):
        """Validation step end."""
        outputs = self.validation_step_outputs
        self.epoch_history["epoch"].append(self.current_epoch)
        self.epoch_history["mode"].append("valid")
        for key in self.epoch_keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))
            self.log(f"val_{key}", self.epoch_history[key][-1], prog_bar=True)

        self.validation_step_outputs.clear()

    def test_step(self, batch, batch_idx):
        """Test step."""
        return self.validation_step(batch, batch_idx)

    def test_epoch_end(self, outputs):
        """Test step end."""
        self.epoch_history["epoch"].append(self.current_epoch)
        self.epoch_history["mode"].append("test")
        for key in self.epoch_keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))
            self.log(f"test_{key}", self.epoch_history[key][-1], prog_bar=True)



"""
Deep Dream recipe.

Performs the algorithm described in
https://ai.googleblog.com/2015/06/inceptionism-going-deeper-into-neural.html

This implementation differs from the original one: the image is optimized in
Fourier space, for greater details and colors, the model and layers are
customizable.

A commandline interface is provided through `python3 -m
torchelie.recipes.deepdream`, and a DeepDreamRecipe is provided.
"""
import random

import torch
import torchvision.transforms as TF

import torchelie.nn as tnn
from torchelie.data_learning import ParameterizedImg
from torchelie.loss.deepdreamloss import DeepDreamLoss
from torchelie.optim import DeepDreamOptim
from torchelie.recipes.recipebase import Recipe
import torchelie.callbacks as tcb

from PIL import Image


class DeepDream(torch.nn.Module):
    """
    Deep Dream recipe

    First instantiate the recipe then call `recipe(n_iter, img)`

    Args:
        model (nn.Module): the trained model to use
        dream_layer (str): the layer to use on which activations will be
            maximized
    """

    def __init__(self, model, dream_layer):
        super(DeepDream, self).__init__()
        self.loss = DeepDreamLoss(model, dream_layer)
        self.norm = tnn.ImageNetInputNorm()

    def fit(self, ref, iters, lr=3e-4, device='cpu', visdom_env='deepdream'):
        """
        Args:
            lr (float, optional): the learning rate
            visdom_env (str or None): the name of the visdom env to use, or None
                to disable Visdom
        """
        ref_tensor = TF.ToTensor()(ref).unsqueeze(0)
        canvas = ParameterizedImg(1, 3,
                                  ref_tensor.shape[2],
                                  ref_tensor.shape[3],
                                  init_img=ref_tensor,
                                  space='spectral',
                                  colors='uncorr')

        def forward(_):
            img = canvas()
            rnd = random.randint(0, 10)
            loss = self.loss(self.norm(img[:, :, rnd:, rnd:]))
            loss.backward()
            return {'loss': loss, 'img': img}

        loop = Recipe(forward, range(iters))
        loop.register('model', self)
        loop.register('canvas', canvas)
        loop.callbacks.add_callbacks([
            tcb.Counter(),
            tcb.Log('loss', 'loss'),
            tcb.Log('img', 'img'),
            tcb.Optimizer(DeepDreamOptim(canvas.parameters(), lr=lr)),
            tcb.VisdomLogger(visdom_env=visdom_env, log_every=10),
            tcb.StdoutLogger(log_every=10)
        ])
        loop.to(device)
        loop.run(1)
        return canvas.render().cpu()


if __name__ == '__main__':
    import argparse
    import torchvision.models as tvmodels

    models = {
        'vgg': {
            'ctor': tvmodels.vgg19,
            'layer': 'features.28'
        },
        'inception': {
            'ctor': tvmodels.inception_v3,
            'layer': 'Mixed_6c'
        },
        'googlenet': {
            'ctor': tvmodels.googlenet,
            'layer': 'inception4c'
        },
        'resnet': {
            'ctor': tvmodels.resnet18,
            'layer': 'layer3'
        }
    }
    parser = argparse.ArgumentParser(description="DeepDream")
    parser.add_argument('--input', required=True)
    parser.add_argument('--out', required=True)
    parser.add_argument('--device', default='cuda')
    parser.add_argument('--model', default='googlenet', choices=models.keys())
    parser.add_argument('--lr', default=3e-4, type=float)
    parser.add_argument('--iters', default=4000, type=int)
    parser.add_argument('--dream-layer')
    parser.add_argument('--visdom-env')
    args = parser.parse_args()

    model = models[args.model]['ctor'](pretrained=True)

    print(model)
    img = Image.open(args.input)
    dd = DeepDream(model, args.dream_layer or models[args.model]['layer'])
    out = dd.fit(img,
                 args.iters,
                 lr=args.lr,
                 device=args.device,
                 visdom_env=args.visdom_env)

    TF.ToPILImage()(out).save(args.out)



import torch
import torch.nn as nn
import torch.nn.functional as F

from torchelie.utils import kaiming, experimental


class MaskedConv2d(nn.Conv2d):
    """
    A masked 2D convolution for PixelCNN

    Args:
        in_chan (int): number of input channels
        out_chan (int): number of output channels
        ks (int): kernel size
        center (bool): whereas central pixel is masked or not
        stride (int): stride, defaults to 1
        bias (2-tuple of ints): A spatial bias. Either the spatial dimensions
            of the input for a different bias at each location, or (1, 1) for
            the same bias everywhere (default)
    """

    def __init__(self, in_chan, out_chan, ks, center, stride=1, bias=(1, 1)):
        super(MaskedConv2d, self).__init__(in_chan,
                                           out_chan, (ks // 2 + 1, ks),
                                           padding=0,
                                           stride=stride,
                                           bias=False)
        self.register_buffer('mask', torch.ones(ks // 2 + 1, ks))
        self.mask[-1, ks // 2 + (1 if center else 0):] = 0

        self.spatial_bias = None
        if bias is not None:
            self.spatial_bias = nn.Parameter(torch.zeros(out_chan, *bias))

        nn.init.kaiming_uniform_(self.weight)

    def forward(self, x):
        self.weight_orig = self.weight
        del self.weight
        self.weight = self.weight_orig * self.mask
        ks = self.weight.shape[-1]

        x = F.pad(x, (ks // 2, ks // 2, ks // 2, 0))
        res = super(MaskedConv2d, self).forward(x)

        self.weight = self.weight_orig
        del self.weight_orig
        if self.spatial_bias is not None:
            return res + self.spatial_bias
        else:
            return res


class TopLeftConv2d(nn.Module):
    """
    A 2D convolution for PixelCNN made of a convolution above the current pixel
    and another on the left.

    Args:
        in_chan (int): number of input channels
        out_chan (int): number of output channels
        ks (int): kernel size
        center (bool): whereas central pixel is masked or not
        stride (int): stride, defaults to 1
        bias (2-tuple of ints): A spatial bias. Either the spatial dimensions
            of the input for a different bias at each location, or (1, 1) for
            the same bias everywhere (default)
    """

    @experimental
    def __init__(self, in_chan, out_chan, ks, center, stride=1, bias=(1, 1)):
        super(TopLeftConv2d, self).__init__()
        self.top = kaiming(
            nn.Conv2d(in_chan,
                      out_chan, (ks // 2, ks),
                      bias=False,
                      stride=stride))
        self.left = kaiming(
            nn.Conv2d(in_chan,
                      out_chan, (1, ks // 2 + (1 if center else 0)),
                      stride=stride,
                      bias=False))
        self.ks = ks
        self.center = center
        self.bias = nn.Parameter(torch.zeros(out_chan, *bias))

    def forward(self, x):
        top = self.top(
            F.pad(x[:, :, :-1, :],
                  (self.ks // 2, self.ks // 2, self.ks // 2, 0)))
        if not self.center:
            left = self.left(F.pad(x[:, :, :, :-1], (self.ks // 2, 0, 0, 0)))
        else:
            left = self.left(F.pad(x, (self.ks // 2, 0, 0, 0)))
        return top + left + self.bias



import torch
from torchvision.transforms.functional import to_tensor
from PIL import Image
from torchvision.datasets.utils import download_and_extract_archive
from typing import Optional, Callable, Tuple, List
import os


class UnlabeledImages:
    """
    Serve all the images contained in a directory and subdirectories without
    any labels and structure constraint.

    Args:
        root (str): path to the root directory
        transform (callable): transformations
    """
    def __init__(self,
                 root: str,
                 transform: Optional[Callable] = None) -> None:
        root = os.path.expanduser(root)
        self.samples = list(
            root + '/' + name for root, _, files in os.walk(root)
            for name in files
            if name.split('.')[-1].lower() in ['bmp', 'jpg', 'jpeg', 'png'])
        self.transform = transform

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, i: int):
        img = Image.open(self.samples[i])
        if self.transform is not None:
            img = self.transform(img)
        return img


class ImagesPaths:
    """
    Serve all the images given in :code:`paths`.

    Args:
        paths (List[str]): paths to images
        transform (callable): transformations
    """
    def __init__(self,
                 paths: List[str],
                 transform: Optional[Callable] = None) -> None:
        self.samples = list(map(os.path.expanduser, paths))
        self.transform = transform

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, i: int):
        img = Image.open(self.samples[i])
        if self.transform is not None:
            img = self.transform(img)
        return img


class SideBySideImagePairsDataset(UnlabeledImages):
    """
    Dataset for side-by-side images. It splits the images so that the same
    transforms are applied to pairs and remain meaningful.
    """
    def __init__(self, root: str, transform: Optional[Callable] = None) -> None:
        super().__init__(root, transform)

    def __getitem__(self, i: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Return the ith paired image as (img1, img2)
        """
        img_path = self.samples[i]
        img = to_tensor(Image.open(img_path))

        img = self._concat(img)

        if self.transform is not None:
            img = self.transform(img)
        return self._split(img)

    def _concat(self, x: torch.Tensor) -> torch.Tensor:
        w = x.shape[2] // 2
        return torch.cat([x[:, :, w:2 * w], x[:, :, :w]], dim=0)

    def _split(self, x):
        c = x.shape[0] // 2
        return x[:c], x[c:]


class Pix2PixDataset(SideBySideImagePairsDataset):
    """
    Paired images datasets made for the Pix2Pix paper for paired image
    translation.

    Args:
        root (str): path to the dataset
        which (str): which dataset to use. One of ['cityscapes',
            'edges2handbags', 'edges2shoes', 'facades', 'maps', 'night2day']
        split (str): choose 'train', 'val' or 'test' set. Default to train.
        download (bool): if True, download the dataset
        transform (optional: function): A callable that transforms the paired
            image represented as a 6 channels tensor.


    cityscapes, maps have train and val splits.

    night2day, facades have traint, test and val splits.
    """
    def __init__(self,
                 root: str,
                 which: str,
                 split: str = 'train',
                 download: bool = False,
                 transform: Optional[Callable] = None) -> None:
        assert which in [
            'cityscapes', 'edges2handbags', 'edges2shoes', 'facades', 'maps',
            'night2day'
        ], f'{which} is not a valid dataset for pix2pix'
        root = root
        if download:
            download_and_extract_archive(
                f'http://efrosgans.eecs.berkeley.edu/pix2pix/datasets/{which}.tar.gz',
                root)
        super().__init__(f'{root}/{which}/{split}', transform=transform)



import torch
import torch.nn as nn
import torchelie.utils as tu
import torchelie.nn as tnn


@tu.experimental
class Hourglass(nn.Module):
    """
    Hourglass model from Deep Image Prior.
    """

    def __init__(self,
                 noise_dim=32,
                 down_channels=[128, 128, 128, 128, 128],
                 skip_channels=4,
                 down_kernel=[3, 3, 3, 3, 3],
                 up_kernel=[3, 3, 3, 3, 3],
                 upsampling='bilinear') -> None:
        super().__init__()

        assert (len(down_channels) == len(down_kernel)), (len(down_channels),
                                                          len(down_kernel))
        assert (len(down_channels) == len(up_kernel)), (len(down_channels),
                                                        len(up_kernel))

        self.upsampling = upsampling
        self.downs = nn.ModuleList(
            [self.down(noise_dim, down_channels[0], down_kernel[0])] + [
                self.down(d1, d2, down_kernel[0]) for d1, d2, k in zip(
                    down_channels[:-1], down_channels[1:], down_kernel[1:])
            ])

        self.ups = nn.ModuleList([
            self.up(down_channels[-1]
                    + skip_channels, down_channels[-1], up_kernel[0])
        ] + [
            self.up(d1 + skip_channels, d2, k) for d1, d2, k in zip(
                down_channels[:0:-1], down_channels[-2::-1], up_kernel[1:])
        ])

        if skip_channels != 0:
            self.skips = nn.ModuleList(
                [self.skip(d, skip_channels) for d in down_channels])

        self.to_rgb = tnn.ConvBlock(down_channels[0], 3, up_kernel[-1])
        self.to_rgb.no_relu()

        tnn.utils.make_leaky(self)
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                m.padding_mode = 'reflect'

    def down(self, in_ch, out_ch, ks) -> nn.Sequential:
        return nn.Sequential(
            tnn.ConvBlock(in_ch, out_ch, ks, stride=2),
            tnn.ConvBlock(out_ch, out_ch, ks, stride=1),
        )

    def up(self, in_ch, out_ch, ks) -> nn.Sequential:
        conv = tnn.ConvBlock(in_ch, out_ch, ks)
        tnn.utils.insert_before(conv, 'conv', nn.BatchNorm2d(in_ch), 'pre_bn')
        return conv

    def skip(self, in_ch, out_ch) -> nn.Sequential:
        return tnn.ConvBlock(in_ch, out_ch, 1)

    def forward(self, x) -> torch.Tensor:
        acts = [x]
        for d in self.downs:
            acts.append(d(acts[-1]))
        acts = acts[1:]

        if hasattr(self, 'skips'):
            skips = [s(a) for s, a in zip(self.skips, acts)]

            x = acts[-1]
            for u, s in zip(self.ups, reversed(skips)):
                x = nn.functional.interpolate(x,
                                              size=s.shape[2:],
                                              mode=self.upsampling)
                x = u(torch.cat([x, s], dim=1))
        else:
            x = acts[-1]
            for u, x2 in zip(self.ups, reversed(acts)):
                x = nn.functional.interpolate(x,
                                              size=x2.shape[2:],
                                              mode=self.upsampling)
                x = u(x)

        x = nn.functional.interpolate(x, scale_factor=2, mode=self.upsampling)
        return torch.sigmoid(self.to_rgb(x))



import torch
import torchvision.transforms as TF
from torch.utils.data import Dataset


class ColoredColumns(Dataset):
    """
    A dataset of precedurally generated images of columns randomly colorized.

    Args:
        *size (int): size of images
        transform (transforms or None): the image transforms to apply to the
            generated pictures
    """
    def __init__(self, *size, transform=None) -> None:
        super(ColoredColumns, self).__init__()
        self.size = size
        self.transform = transform if transform is not None else (lambda x: x)

    def __len__(self):
        return 10000

    def __getitem__(self, i):
        cols = torch.randint(0, 255, (3, 1, self.size[1]))
        expanded = cols.expand(3, *self.size).float()
        img = TF.ToPILImage()(expanded / 255)
        return self.transform(img), 0


class ColoredRows(Dataset):
    """
    A dataset of precedurally generated images of rows randomly colorized.

    Args:
        *size (int): size of images
        transform (transforms or None): the image transforms to apply to the
            generated pictures
    """
    def __init__(self, *size, transform=None) -> None:
        super(ColoredRows, self).__init__()
        self.size = size
        self.transform = transform if transform is not None else (lambda x: x)

    def __len__(self):
        return 10000

    def __getitem__(self, i):
        rows = torch.randint(0, 255, (3, self.size[0], 1))
        expanded = rows.expand(3, *self.size).float()
        img = TF.ToPILImage()(expanded / 255)
        return self.transform(img), 0



from .reshape import Reshape, Lambda
from .conv import *
from .debug import Debug, Dummy
from .noise import Noise
from .vq import VQ, MultiVQ
from .imagenetinputnorm import ImageNetInputNorm
from .withsavedactivations import WithSavedActivations
from .maskedconv import MaskedConv2d, TopLeftConv2d
from .batchnorm import *
from .adain import AdaIN2d, FiLM2d
from .pixelnorm import PixelNorm
from .blocks import *
from .layers import *
from .condseq import CondSeq
from .graph import ModuleGraph
from .encdec import *
from .interpolate import *
from .resblock import *
import torchelie.nn.utils



import torch
import torch.nn as nn
import torchelie.nn as tnn
import torchelie.utils as tu
from typing import List, Union, Optional
from .classifier import ClassificationHead, ProjectionDiscr


class ResidualDiscriminator(nn.Module):
    classifier: Union[ClassificationHead, ProjectionDiscr]

    def __init__(self, arch: List[Union[str, int]]) -> None:
        super().__init__()
        self.arch = arch
        self.in_channels = 3
        in_ch = arch[0]

        features = tnn.CondSeq()
        assert isinstance(in_ch, int)
        features.add_module('input', tnn.Conv3x3(3, in_ch))

        ii = 0
        for i, (x, x2) in enumerate(zip(arch, arch[1:] + ['dummy'])):
            if x == 'D':
                continue

            downsample = x2 == 'D'
            assert isinstance(x, int)
            features.add_module(f'block_{ii}',
                                tnn.ResidualDiscrBlock(in_ch, x, downsample))
            in_ch = x
            ii += 1
        self.out_channels = in_ch
        features.add_module('final_relu', nn.LeakyReLU(0.2, True))
        assert isinstance(features.block_0, tnn.ResidualDiscrBlock)
        features.block_0.preact_skip()
        self.features = features

        self.classifier = ClassificationHead(self.out_channels, 1)

    def forward(self,
                x: torch.Tensor,
                y: Optional[torch.Tensor] = None) -> torch.Tensor:
        x = self.features(x)
        if isinstance(self.classifier, ProjectionDiscr):
            return self.classifier(x, y)
        else:
            return self.classifier(x)

    def set_input_specs(self, in_channels: int) -> 'ResidualDiscriminator':
        self.features.input = tnn.Conv3x3(in_channels,
                                          self.features.input.out_channels)
        return self

    @tu.experimental
    def to_spectral_norm(self) -> 'ResidualDiscriminator':
        for m in self.modules():
            if isinstance(m, tnn.ResidualDiscrBlock):
                m.to_spectral_norm()
            elif isinstance(m, nn.Linear):
                nn.utils.spectral_norm(m)

        assert isinstance(self.features.input, nn.Module)
        nn.utils.spectral_norm(self.features.input)
        if hasattr(self.features, 'mbconv'):
            assert isinstance(self.features.mbconv, tnn.ConvBlock)
            assert isinstance(self.features.mbconv.conv, nn.Module)
            nn.utils.spectral_norm(self.features.mbconv.conv)
        if isinstance(self.classifier, ProjectionDiscr):
            self.classifier.to_spectral_norm()
        return self

    @tu.experimental
    def to_equal_lr(self, leak=0.2) -> 'ResidualDiscriminator':
        for m in self.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                tu.kaiming(m, dynamic=True, a=leak)

        return self

    def add_minibatch_stddev(self) -> 'ResidualDiscriminator':
        out_ch = self.out_channels
        self.features.add_module('mbstd', tnn.MinibatchStddev())
        self.features.add_module(
            'mbconv',
            tnn.ConvBlock(out_ch + 1, out_ch, 3).remove_batchnorm().leaky())
        return self

    def to_projection_discr(self, num_classes: int) -> 'ResidualDiscriminator':
        self.classifier = ProjectionDiscr(self.out_channels, num_classes)
        return self


def residual_patch34():
    D = ResidualDiscriminator([32, 'D', 64, 'D', 128])
    D.classifier.to_convolutional().leaky()
    return D


def residual_patch70():
    D = ResidualDiscriminator([32, 'D', 64, 'D', 128, 'D', 256])
    D.classifier.to_convolutional().leaky()
    return D


def residual_patch142():
    D = ResidualDiscriminator([32, 'D', 64, 'D', 128, 'D', 256, 'D', 512])
    D.classifier.to_convolutional().leaky()
    return D


def residual_patch286():
    D = ResidualDiscriminator(
        [32, 'D', 64, 'D', 128, 'D', 256, 'D', 512, 'D', 512])
    D.classifier.to_convolutional().leaky()
    return D


def res_discr_3l() -> ResidualDiscriminator:
    return ResidualDiscriminator([32, 'D', 64, 'D', 128, 'D'])


def res_discr_4l() -> ResidualDiscriminator:
    return ResidualDiscriminator([32, 'D', 64, 'D', 128, 'D', 256, 'D'])


def snres_discr_4l() -> ResidualDiscriminator:
    return res_discr_4l().to_spectral_norm()


def snres_projdiscr_4l(num_classes: int) -> ResidualDiscriminator:
    return res_discr_4l().to_projection_discr(num_classes).to_spectral_norm()


def res_discr_5l() -> ResidualDiscriminator:
    return ResidualDiscriminator(
        [32, 'D', 64, 'D', 128, 'D', 256, 'D', 512, 'D'])


def snres_discr_5l() -> ResidualDiscriminator:
    return res_discr_5l().to_spectral_norm()


def snres_projdiscr_5l(num_classes: int) -> ResidualDiscriminator:
    return res_discr_5l().to_projection_discr(num_classes).to_spectral_norm()


def res_discr_6l() -> ResidualDiscriminator:
    return ResidualDiscriminator(
        [32, 'D', 64, 'D', 128, 'D', 256, 'D', 512, 'D', 512, 'D'])


def snres_discr_6l() -> ResidualDiscriminator:
    return res_discr_6l().to_spectral_norm()


def snres_projdiscr_6l(num_classes: int) -> ResidualDiscriminator:
    return res_discr_6l().to_projection_discr(num_classes).to_spectral_norm()


def res_discr_7l() -> ResidualDiscriminator:
    return ResidualDiscriminator(
        [32, 'D', 64, 'D', 128, 'D', 256, 'D', 512, 'D', 512, 'D', 512, 'D'])


def snres_discr_7l() -> ResidualDiscriminator:
    return res_discr_7l().to_spectral_norm()


def snres_projdiscr_7l(num_classes: int) -> ResidualDiscriminator:
    return res_discr_7l().to_projection_discr(num_classes).to_spectral_norm()



import math

import torch
import torchvision.transforms as TF
import torchvision.transforms.functional as F
from PIL import Image
import numpy as np

try:
    import cv2
    HAS_CV = True
except:
    pass


class ResizeNoCrop:
    """
    Resize a PIL image so that its longer border is of size `size`

    Args:
        size (int): max size of the image
    """

    def __init__(self, size, mode=Image.BILINEAR):
        self.size = size
        self.mode = mode

    def __call__(self, x):
        """
        Args:
            x (PIL.Image): the image to transform

        Returns:
            resized image
        """
        size = self.size
        width = x.width
        height = x.height
        if width > height:
            return x.resize((size, int(size * height / width)), self.mode)
        else:
            return x.resize((int(size * width / height), size), self.mode)


class AdaptPad:
    """
    Pad an input image so that it reaches size `size`

    Args:
        sz ((int, int)): target size
        padding_mode (str): one of the modes of `torchvision.transforms.pad`
    """

    def __init__(self, sz, padding_mode='constant', fill=0):
        self.sz = sz
        self.padding_mode = padding_mode
        self.fill = fill

    def __call__(self, img):
        """
        Pad the image

        Args:
            img (PIL.Image): The image to pad

        Returns:
            Padded image
        """
        w, h = img.width, img.height

        pl = max(0, self.sz[1] - w) // 2
        pr = self.sz[1] - w - pl

        pt = max(0, self.sz[0] - h) // 2
        pd = self.sz[0] - h - pt

        return TF.functional.pad(img, (pl, pt, pr, pd),
                                 padding_mode=self.padding_mode,
                                 fill=self.fill)


class MultiBranch:
    """
    Transform an image with multiple transforms

    Args:
        transforms (list of transforms): the parallel set of transforms
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, x):
        """
        Transform the image

        Args:
            x: image

        Returns:
            A tuple `out` so that `out[i] = transforms[i](x)`
        """
        return tuple(tf(x) for tf in self.transforms)


class Canny:
    """
    Run Canny edge detector over an image. Requires OpenCV to be installed

    Args:
        thresh_low (int): lower threshold (default: 100)
        thresh_high (int): upper threshold (default: 200)
    """

    def __init__(self, thresh_low=100, thresh_high=200):
        self.thresh_low = thresh_low
        self.thresh_high = thresh_high

    def __call__(self, img):
        """
        Detect edges

        Args:
            img (PIL.Image): the image

        Returns:
            edges detected in `img` as PIL Image
        """
        assert HAS_CV, ("Can't import OpenCV. Some transforms will not "
                        "work properly")
        img = np.array(img)
        img = cv2.Canny(img, self.thresh_low, self.thresh_high)
        return Image.fromarray(img)


class ResizedCrop(object):
    """Crop the given PIL Image to size.
    A crop of size of the original size is made. This crop
    is finally resized to given size.

    Args:
        size: expected output size of each edge
        scale: size of the origin size cropped
        interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, scale=0.54, ratio=1,
                 interpolation=Image.BILINEAR):
        if isinstance(size, tuple):
            self.size = size
        else:
            self.size = (size, size)

        self.ratio = ratio
        self.interpolation = interpolation
        self.scale = scale

    @staticmethod
    def _get_image_size(img):
        if F._is_pil_image(img):
            return img.size
        elif isinstance(img, torch.Tensor) and img.dim() > 2:
            return img.shape[-2:][::-1]
        else:
            raise TypeError("Unexpected type {}".format(type(img)))

    @staticmethod
    def get_params(img, scale, ratio=1):
        """Get parameters for ``crop``.
        Args:
            img (PIL Image): Image to be cropped.
            scale (float): range of size of the origin size cropped
        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop``.
        """
        scale = math.sqrt(scale)
        width, height = ResizedCrop._get_image_size(img)

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if (in_ratio < 1):
            w = width * scale
            h = int(round(w / ratio))
        elif (in_ratio > 1):
            h = height * scale
            w = int(round(h * ratio))
        else:  # whole image
            w = width * scale
            h = height * scale
        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be cropped and resized.
        Returns:
            PIL Image: Randomly cropped and resized image.
        """
        i, j, h, w = self.get_params(img, self.scale, self.ratio)
        return F.resized_crop(img, i, j, h, w, self.size, self.interpolation)

    def __repr__(self):
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(round(self.scale, 4))
        format_string += ', interpolation={0})'.format(self.interpolation)
        return format_string


class Noise:
    def __init__(self, std):
        self.std = std

    def __call__(self, x):
        return x + torch.randn_like(x) * self.std


def patches(img, patch_size=128):
    """
    Cut an image into square patches of equal size. Padding is added if needed.

    Args:
        img (PIL.Image): the image to split
        patch_size (int): the size od the square patch

    Returns:
        A list of split and coordinates like [((y pos, x pos), img patch)]
    """
    h, w = img.height, img.width
    padder = AdaptPad((patch_size, patch_size))
    for row, h_off in enumerate(range(0, h, patch_size)):
        for c, w_off in enumerate(range(0, w, patch_size)):
            p = img.crop(
                (w_off, h_off, w_off + patch_size, h_off + patch_size))
            yield (row, c), padder(p)


def paste_patches(patches):
    """
    Collate a list of patches and their coordinates such as returned by
    `patches` into an image.

    Args:
        patches (list): a list of patches and coordinates such as returned by
            `patches`

    Returns:
        the assembled PIL.Image.
    """
    patch_size = patches[0][1].width
    total_height = patch_size * (1 + max(p[0][0] for p in patches))
    total_width = patch_size * (1 + max(p[0][1] for p in patches))
    new_im = Image.new('RGB', (total_width, total_height))
    for (l, c), p in patches:
        new_im.paste(p, (c * patch_size, l * patch_size))
    return new_im



import torch

import torchelie.callbacks as tcb
from torchelie.recipes import Recipe


def TrainAndTest(model,
                 train_fun,
                 test_fun,
                 train_loader,
                 test_loader,
                 *,
                 test_every=100,
                 visdom_env='main',
                 log_every=10,
                 checkpoint='model',
                 key_best=None):
    """
    Two nested loops, usually one for training and one for testing, but can
    serve other purposes. The model is automatically registered and
    checkpointed as :code:`checkpoint['model']`, and put in eval mode when
    testing.

    Training callbacks:

    - Counter for counting iterations, connected to the testing loop as well
    - VisdomLogger
    - StdoutLogger

    Testing:

    Testing loop is in :code:`.test_loop`.

    Testing callbacks:

    - VisdomLogger
    - StdoutLogger
    - Checkpoint

    Args:
        model (nn.Model): a model
        train_fun (Callabble): a function that takes a batch as a single
            argument, performs a training step and return a dict of values to
            populate the recipe's state.
        test_fun (Callable): a function taking a batch as a single argument
            then performs something to evaluate your model and returns a dict
            to populate the state.
        train_loader (DataLoader): Training set dataloader
        test_loader (DataLoader): Testing set dataloader
        test_every (int): testing frequency, in number of iterations (default:
            100)
        visdom_env (str): name of the visdom environment to use, or None for
            not using Visdom (default: None)
        log_every (int): logging frequency, in number of iterations (default:
            100)
        checkpoint (str): checkpointing path or None for no checkpointing
        key_best (function or None): a key function for comparing states.
            Checkpointing the greatest.
    """

    def eval_call(batch):
        model.eval()
        with torch.no_grad():
            out = test_fun(batch)
        model.train()
        return out

    train_loop = Recipe(train_fun, train_loader)
    train_loop.register('model', model)

    test_loop = Recipe(eval_call, test_loader)
    train_loop.test_loop = test_loop
    train_loop.register('test_loop', test_loop)

    def prepare_test(state):
        test_loop.callbacks.update_state({
            'epoch': state['epoch'],
            'iters': state['iters'],
            'epoch_batch': state['epoch_batch']
        })

    train_loop.callbacks.add_prologues([tcb.Counter()])
    train_loop.callbacks.add_epilogues([
        tcb.CallRecipe(test_loop, test_every, init_fun=prepare_test),
        tcb.VisdomLogger(visdom_env=visdom_env, log_every=log_every),
        tcb.StdoutLogger(log_every=log_every),
    ])

    test_loop.callbacks.add_epilogues([
        tcb.VisdomLogger(visdom_env=visdom_env, log_every=-1, prefix='test_'),
        tcb.StdoutLogger(log_every=-1, prefix='Test'),
    ])

    if checkpoint is not None:
        test_loop.callbacks.add_epilogues([
            tcb.Checkpoint(checkpoint + '/ckpt_{iters}.pth',
                           train_loop,
                           key_best=key_best)
        ])

    return train_loop



"""
Hyper parameters related utils. For now it contains hyper parameter search
tools with random search but it may evolve later.

::

    hp_sampler = HyperparamSearch(
        lr=ExpSampler(1e-6, 0.1),
        momentum=DecaySampler(0.5, 0.999),
        wd=ExpSampler(0.1, 1e-6),
        model=ChoiceSampler(['resnet', 'vgg'])
    )

    for run_nb in range(10):
        hps = hp_sampler.sample()
        results = train(**hps)
        hp_sampler.log_results(hps, results)


After creating the json file summing up the hyper parameter search, it can be
investigated with the viewer with
:code:`python3 -m torchelie.hyper hpsearch.json`. Then locate your browser to
:code:`http://localhost:8080`.
"""
import math
import random
import json
import copy


class Sampler:
    """
    Uniform sampler.

    Args:
        low (float): lower bound
        high (float): higher bound
    """
    def __init__(self, low: float, high: float) -> None:
        self.low = low
        self.high = high

    def sample(self) -> float:
        """
        Sample from the distribution.
        """
        return random.uniform(self.low, self.high)

    def inverse(self, x: float) -> float:
        return x


class ExpSampler(Sampler):
    """
    Exponential sampler (Uniform sampler over a log scale). Use it to sample
    the learning rate or the weight decay.

    Args:
        low (float): lower bound
        high (float): higher bound
    """
    def __init__(self, low, high):
        low = self.inverse(low)
        high = self.inverse(high)
        super(ExpSampler, self).__init__(low, high)

    def sample(self):
        """
        Sample a value.
        """
        return 10**super(ExpSampler, self).sample()

    def inverse(self, x):
        return math.log10(x)


class DecaySampler(ExpSampler):
    """
    Sample a decay value. Use it for a momentum or beta1 / beta2 value or any
    exponential decay value.

    Args:
        low (float): lower bound
        high (float): higher bound
    """
    def __init__(self, low, high):
        super(DecaySampler, self).__init__(low, high)

    def sample(self):
        """
        Sample a value.
        """
        return 1 - super(DecaySampler, self).sample()

    def inverse(self, x):
        return super(DecaySampler, self).inverse(1 - x)


class ChoiceSampler:
    """
    Sampler over a discrete of values.

    Args:
        choices (list): list of values
    """
    def __init__(self, choices):
        self.choices = choices

    def sample(self):
        """
        Sample a value
        """
        return random.choice(self.choices)

    def inverse(self, x):
        return self.choices.index(x)


class HyperparamSampler:
    """
    Sample hyper parameters. It aggregates multiple samplers to produce a set
    of hyper parameters.

    Example:

    ::
        HyperparamSampler(
            lr=ExpSampler(1e-6, 0.1),
            momentum=DecaySampler(0.5, 0.999),
            wd=ExpSampler(0.1, 1e-6),
            model=ChoiceSampler(['resnet', 'vgg'])
        )

    Args:
        hyperparams (kwargs): hyper params samplers. Names are arbitrary.
    """
    def __init__(self, **hyperparams):
        self.hyperparams = hyperparams

    def sample(self):
        """
        Sample hyperparameters.

        Returns:
            a dict containing sampled values for hyper parameters.
        """
        return {k: v.sample() for k, v in self.hyperparams.items()}


class HyperparamSearch:
    """
    Perform hyper parameter search. Right now it just uses a random search.
    Params and results are logged to hpsearch.csv in the current directory. It
    would be cool to implement something like a Gaussian Process search or a RL
    algorithm.

    First, call sample() to get a set of hyper parameters. The, evaluate them
    on your task and get a dict of results. call log_results() with the hyper
    params and the results dict, and start again. Stop whenever you want.

    Args:
        hyperparameters (kwargs): named samplers (like for HyperparamSampler).
    """
    def __init__(self, **hyperparams):
        self.sampler = HyperparamSampler(**hyperparams)

    def sample(self):
        """
        Sample a set of hyper parameters.
        """
        return self.sampler.sample()

    def read_hpsearch(self):
        try:
            with open('hpsearch.json', 'r') as f:
                return json.load(f)
        except:
            return []

    def _str(self, x):
        try:
            return float(x)
        except:
            pass

        try:
            return x.item()
        except:
            pass

        return x

    def log_result(self, hps, result):
        """
        Logs hyper parameters and results.
        """
        res = self.read_hpsearch()
        full = copy.deepcopy(hps)
        full.update({'result_' + k: self._str(v) for k, v in result.items()})
        res.append(full)
        with open('hpsearch.json', 'w') as f:
            json.dump(res, f)


if __name__ == '__main__':
    from http.server import HTTPServer, BaseHTTPRequestHandler
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('file', default='hpsearch.json')
    opts = parser.parse_args()

    def make_html():
        with open(opts.file) as f:
            dat = json.load(f)

        dat.sort(key=lambda x: x['result_lfw_loss'])
        dat = dat
        dimensions = []
        for k in dat[0].keys():
            v = dat[0][k]
            if isinstance(v, float):
                dimensions.append({
                    'label': k,
                    'values': [dd[k] for dd in dat]
                })
        print(json.dumps(dimensions))
        return """
        <html>
            <head>
                <meta charset="UTF-8">
                <title></title>
                <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
                <script
                src="https://cdn.jsdelivr.net/npm/sorttable@1.0.2/sorttable.min.js"></script>
            <style>
            table, th, td {
              border: 1px solid black;
            }
            </style>
            </head>
            <body>
                <div id="graphOpts"></div>
                <div id="graphDiv"></div>
                <div id="table"></div>
        <script>
        var DATA = """ + json.dumps(dat) + """
        for (d in DATA) {
            DATA[d]['idx'] = d
        }

        let display_opts = {};
        let hidden_rows = new Set();
        function make_graph() {
            let data = DATA.filter((_, i) => !hidden_rows.has(i));
            let DIM = Object.keys(data[0])
                .filter(k => display_opts[k] !== "Hidden")
                .map(k => {
                    if (typeof data[0][k] == 'number') {
                        if (display_opts[k] == 'Visible'
                                || display_opts[k] == undefined) {
                            return {
                                'label': k,
                                'values': data.map(d => d[k])
                            };
                        } else if (display_opts[k] == 'Log') {
                            return {
                                'label': k,
                                'values': data.map(d => Math.log10(d[k])),
                                'tickvals': data.map(d => Math.log10(d[k])),
                                'ticktext': data.map(d => d[k].toExponential(2))
                            };
                        } else if (display_opts[k] == 'Decay') {
                            return {
                                'label': k,
                                'values': data.map(d => Math.log10(1/(1-d[k]))),
                                'tickvals': data.map(d => Math.log10(1/(1-d[k]))),
                                'ticktext': data.map(d => d[k].toExponential(2))
                            };
                        }
                    } else {
                        let labels = Array.from(new Set(data.map(d => d[k])));

                        return {
                            'label': k,
                            'values': data.map(d => labels.indexOf(d[k])),
                            'tickvals': labels.map((_, i) => i),
                            'ticktext': labels
                        };
                    }
                });
            var trace = {
              type: 'parcoords',
              line: {
                color: 'blue'
              },

              dimensions: DIM
            };

            var show = [trace]

            Plotly.purge('graphDiv');
            Plotly.plot('graphDiv', show, {}, {showSendToCloud: true});
        }
        make_graph();

        function change_opts(t, val) {
            display_opts[t]=val;
            make_graph();
        }

        graphOpts.innerHTML = `
            <table width="100%">
                <tr>
                    ${get_titles().map(t => `<th>${t}</th>`).join('')}
                </tr>
                <tr>
                    ${get_titles().map(t =>
                        `<td>
                            <select onchange="change_opts('${t}', event.target.value)">
                                <option default>Visible</option>
                                <option>Hidden</option>
                                <option>Log</option>
                                <option>Decay</option>
                            </select>
                        </th>`
                    ).join('')}
                </tr>
            </table>
        `;

        function get_titles() {
            let titles = [];
            for (d of DATA) {
                for (k of Object.keys(d)) {
                    titles.push(k);
                }
            }
            titles = Array.from(new Set(titles));
            return titles;
        }

        function change_visibility(r, hid) {
            if (hid) {
                hidden_rows.add(r);
            } else {
                hidden_rows.delete(r);
            }
            make_graph();
        }
        let make_table = (DATA) => {
            let titles = get_titles();
            return `
                <table width="100%" class="sortable">
                    <tr>
                    ${titles.map(t =>
                        `<th>${t}</th>`
                    ).join('')}
                    <th>Hide</th>
                    </tr>
                ${DATA.map((d, i) =>
                    `<tr>${titles.map(t =>
                        `<td>
                            ${(typeof d[t] == 'number')
                                ? d[t].toFixed(3)
                                : d[t]}
                        </td>`
                        ).join('')}
                        <td>H
                            <input type="checkbox"
                            onchange="change_visibility(${i}, event.target.checked)"/>
                        </td>
                    </tr>`).join('')}
            </table>`;
        };
        table.innerHTML = make_table(DATA);
        </script>
        </body>
        </html>
        """

    class Server(BaseHTTPRequestHandler):
        def do_GET(self):
            self.handle_http(200, 'text/html', make_html())

        def handle_http(self, status, content_type, content):
            self.send_response(status)
            self.send_header('Content-type', content_type)
            self.end_headers()
            self.wfile.write(bytes(content, 'utf-8'))

    httpd = HTTPServer(('0.0.0.0', 8080), Server)
    try:
        print('serving on localhost:8080')
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()



"""
This example demonstrates how to learn MNIST with Torchelie. It can be
trivially modified to fit another dataset or model.

Better than that, make sure to check the Classification Recipe's builtin
command line interface that allows to fit a model to an image dataset without
writing a single line of code. It is good to quickly estimate how hard a
dataset is to learn by fitting a default model with default transforms and
hyperparameters.
"""
import argparse

import torch
import torch.nn as nn
import torch.nn.functional as F

from torchvision.datasets import CIFAR10
import torchvision.transforms as TF

import torchelie as tch
from torchelie.recipes.classification import CrossEntropyClassification


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--device', choices=['cpu', 'cuda'], default='cpu')
    parser.add_argument('--epochs', default=1, type=int)
    parser.add_argument('--lr', type=float, default=1e-3)
    return parser.parse_args()


def build_transforms():
    tfms = TF.Compose([
        TF.Resize(32),
        TF.ToTensor(),
        TF.Normalize([0.5] * 3, [0.5] * 3, True),
    ])
    train_tfms = TF.Compose([
        TF.RandomCrop(32, padding=4),
        TF.RandomHorizontalFlip(),
        TF.ToTensor(),
        TF.Normalize([0.5] * 3, [0.5] * 3, True),
    ])
    return tfms, train_tfms


def get_datasets():
    tfms, train_tfms = build_transforms()
    ds = CIFAR10('~/.cache/torch/cifar10', download=True, transform=train_tfms)
    dst = CIFAR10('~/.cache/torch/cifar10',
                  transform=tfms,
                  download=True,
                  train=False)
    return ds, dst


def train():
    opts = get_args()
    ds, dst = get_datasets()
    model = tch.models.preact_resnet20_cifar(num_classes=10)
    print(model)
    dl = torch.utils.data.DataLoader(ds,
                                     num_workers=4,
                                     batch_size=128,
                                     pin_memory=True,
                                     shuffle=True)
    dlt = torch.utils.data.DataLoader(dst,
                                      num_workers=4,
                                      batch_size=256,
                                      pin_memory=True
                                      )
    recipe = CrossEntropyClassification(model,
                                        dl,
                                        dlt,
                                        ds.classes,
                                        lr=opts.lr,
                                        wd=0.1,
                                        beta1=0.9,
                                        log_every=100,
                                        test_every=500,
                                        visdom_env='cifar_preactresnet20')

    recipe.to(opts.device)
    recipe.run(opts.epochs)


if __name__ == '__main__':
    train()



from io import BytesIO
import base64 as b64
from typing import Union, List, Optional

import torch
from torchelie.utils import as_multiclass_shape, experimental

import numpy as np
from PIL import Image


def img2html(img: Union[torch.Tensor, np.ndarray], opts: dict = None) -> str:
    """
    Convert an image to a b64 html inline image.

    Possible options: width (int, pixels), height (int, pixels) or jpgquality
    (int, percentage)
    """
    if isinstance(img, torch.Tensor):
        img = img.cpu().detach().numpy()

    opts = {} if opts is None else opts
    if isinstance(img, np.ndarray):
        nchannels = img.shape[0] if img.ndim == 3 else 1
        if nchannels == 1:
            img = np.squeeze(img)
            img = img[np.newaxis, :, :].repeat(3, axis=0)

        if 'float' in str(img.dtype):
            if img.max() <= 1:
                img = img * 255.
            img = np.uint8(img)

        img = np.transpose(img, (1, 2, 0))
        img = Image.fromarray(img)

    opts['width'] = opts.get('width', img.width)
    opts['height'] = opts.get('height', img.height)

    buf = BytesIO()
    image_type = 'png'
    imsave_args = {}
    if 'jpgquality' in opts:
        image_type = 'jpeg'
        imsave_args['quality'] = opts['jpgquality']

    img.save(buf, format=image_type.upper(), **imsave_args)

    b64encoded = b64.b64encode(buf.getvalue()).decode('utf-8')

    return '<img src="data:image/{};base64,{}"/>'.format(
        image_type, b64encoded)


class ClassificationInspector:
    """
    Visdom HTML display of classification results, with best, worst, and mode
    indecisive results.
    """
    def __init__(self,
                 topk: int,
                 labels: List[str],
                 center_value: float = 0) -> None:
        self.labels = labels
        self.center_value = center_value
        self.topk = topk
        self.reset()
        self.m = float('inf')
        self.M = float('-inf')

    def reset(self):
        self.best = []
        self.worst = []
        self.confused = []

    def analyze(self,
                batch: torch.Tensor,
                pred: torch.Tensor,
                true: torch.Tensor,
                pred_label: Optional[torch.Tensor] = None,
                paths: Optional[List[str]] = None) -> None:
        pred = as_multiclass_shape(pred, as_probs=True)
        for_label = pred.gather(1, true.unsqueeze(1))
        if pred_label is None:
            pred_label = pred.argmax(dim=1)
        best_pred = pred.gather(1, pred_label.unsqueeze(1))
        if paths is None:
            paths = [None] * len(batch)
        this_data = list(
            zip(batch, for_label, true, pred_label == true, paths, pred_label,
                best_pred))

        self.best += this_data
        self.best.sort(key=lambda x: -x[1])
        self.best = self.best[:self.topk]

        self.worst += this_data
        self.worst.sort(key=lambda x: x[1])
        self.worst = self.worst[:self.topk]

        self.confused += this_data
        self.confused.sort(key=lambda x: abs(self.center_value - x[1]))
        self.confused = self.confused[:self.topk]

    def _report(self, dat):
        def prob_as_bar(cos, is_correct):
            return '<div style="width:{percents}%;background-color:{color};height:5px"></div>'.format(
                percents=int(cos * 100),
                color='green' if is_correct else 'red')

        html = ['<div style="display:flex;flex-wrap:wrap">']
        for img, p, cls, correct, path, pred_label, best_pred in dat:
            self.m = min(self.m, img.min())
            img = img - self.m
            self.M = max(self.M, img.max())
            img /= self.M

            tlabel = self.labels[cls.item()]
            tlabel = tlabel.replace('_', ' ').replace('-', ' ')

            plabel = self.labels[pred_label.item()]
            plabel = plabel.replace('_', ' ').replace('-', ' ')
            html.append(
                ('<div onclick="javascript:prompt(\'path\', \'{path}\')">'
                 '<div style="padding:3px;width:{px_sz}px">'
                 '{img}{bar}{checkmark}{true_label} (pred: {pred_label})'
                 '</div>'
                 '</div>').format(path=path,
                                  px_sz=dat[0][0].shape[2],
                                  img=img2html(img),
                                  bar=prob_as_bar(best_pred.item(),
                                                  correct.item()),
                                  checkmark='✓' if correct.item() else '✗',
                                  true_label=tlabel,
                                  pred_label=plabel))
        html.append('</div>')
        return ''.join(html)

    def show(self) -> str:
        """
        Get the HTML inspector view.
        """
        html = [
            '<h1>Best predictions</h1>',
            self._report(self.best), '<h1>Worst predictions</h1>',
            self._report(self.worst), '<h1>Confusions</h1>',
            self._report(self.confused)
        ]
        return ''.join(html)


class SegmentationInspector(ClassificationInspector):
    @experimental
    def __init__(self, topk, labels, center_value=0):
        super().__init__(topk, labels, center_value=0)

    def analyze(self,
                batch: torch.Tensor,
                pred: torch.Tensor,
                true: torch.Tensor,
                pred_label: Optional[torch.Tensor] = None,
                paths: Optional[List[str]] = None) -> None:
        pred = torch.sigmoid(pred)
        for_label = torch.median(
            (pred * true + (1 - pred) * (1 - true)).reshape(pred.shape[0],
                                                            -1), -1)[0]
        if pred_label is None:
            pred_label = (pred > 0.5).int()
        if paths is None:
            paths = [None] * len(batch)
        this_data = list(
            zip(batch, for_label,
                true.mean(tuple(range(1, true.dim()))) > 0.5,
                (pred_label == true).float().mean(
                    tuple(range(1, pred_label.dim()))) > 0.5, paths))

        self.best += this_data
        self.best.sort(key=lambda x: -x[1])
        self.best = self.best[:self.topk]

        self.worst += this_data
        self.worst.sort(key=lambda x: x[1])
        self.worst = self.worst[:self.topk]

        self.confused += this_data
        self.confused.sort(key=lambda x: abs(self.center_value - x[1]))
        self.confused = self.confused[:self.topk]



"""
Neural Style from Leon Gatys

A commandline interface is provided through `python3 -m
torchelie.recipes.neural_style`
"""

import torch
from torchvision.transforms.functional import to_pil_image, to_tensor

import torchelie as tch
from torchelie.loss import NeuralStyleLoss
from torchelie.data_learning import ParameterizedImg
from torchelie.recipes.recipebase import Recipe
import torchelie.callbacks as tcb


class NeuralStyle(torch.nn.Module):
    """
    Neural Style Recipe

    First instantiate the recipe then call `recipe(n_iter, img)`

    Args:
        lr (float, optional): the learning rate
        device (device): where to run the computation
        visdom_env (str or None): the name of the visdom env to use, or None
            to disable Visdom
    """

    def __init__(self, lr=0.01, device="cpu", visdom_env='style'):
        super(NeuralStyle, self).__init__()
        self.loss = NeuralStyleLoss()
        self.loss2 = NeuralStyleLoss()
        self.device = device
        self.lr = lr
        self.visdom_env = visdom_env

    def fit(self,
            iters,
            content_img,
            style_img,
            style_ratio,
            *,
            second_scale_ratio=1,
            content_layers=None,
            init_with_content=False):
        """
        Run the recipe

        Args:
            n_iters (int): number of iterations to run
            content (PIL.Image): content image
            style (PIL.Image): style image
            ratio (float): weight of style loss
            content_layers (list of str): layers on which to reconstruct
                content
        """
        self.loss.to(self.device)
        self.loss.set_style(to_tensor(style_img).to(self.device), style_ratio)
        self.loss.set_content(
            to_tensor(content_img).to(self.device), content_layers)

        self.loss2.to(self.device)
        self.loss2.set_style(
            torch.nn.functional.interpolate(to_tensor(style_img)[None],
                                            scale_factor=0.5,
                                            mode='bilinear',
                                            align_corners=False,
                                            recompute_scale_factor=True)[0].to(
                                                self.device), style_ratio)
        self.loss2.set_content(
            torch.nn.functional.interpolate(to_tensor(content_img)[None],
                                            scale_factor=0.5,
                                            mode='bilinear',
                                            align_corners=False,
                                            recompute_scale_factor=True)[0].to(
                                                self.device), content_layers)

        canvas = ParameterizedImg(1,
                                  3,
                                  content_img.height,
                                  content_img.width,
                                  init_img=to_tensor(content_img).unsqueeze(0)
                                  if init_with_content else None)

        self.opt = tch.optim.RAdamW(canvas.parameters(),
                                    3e-2,
                                    betas=(0.9, 0.99),
                                    weight_decay=0)

        def forward(_):
            img = canvas()
            loss, losses = self.loss(img)
            loss.backward()
            loss, losses = self.loss2(
                torch.nn.functional.interpolate(canvas(),
                                                scale_factor=0.5,
                                                mode='bilinear',
                                                align_corners=False,
                                                recompute_scale_factor=True))
            (second_scale_ratio * loss).backward()

            return {
                'loss': loss,
                'content_loss': losses['content_loss'],
                'style_loss': losses['style_loss'],
                'hists_loss': losses['hists_loss'],
                'img': img,
            }

        loop = Recipe(forward, range(iters))
        loop.register('canvas', canvas)
        loop.register('model', self)
        loop.callbacks.add_callbacks([
            tcb.Counter(),
            tcb.WindowedMetricAvg('loss'),
            tcb.WindowedMetricAvg('content_loss'),
            tcb.WindowedMetricAvg('style_loss'),
            tcb.WindowedMetricAvg('hists_loss'),
            tcb.Log('img', 'img'),
            tcb.VisdomLogger(visdom_env=self.visdom_env, log_every=10),
            tcb.StdoutLogger(log_every=10),
            tcb.Optimizer(self.opt, log_lr=True),
        ])
        loop.to(self.device)
        loop.run(1)
        return canvas.render().cpu()


if __name__ == '__main__':
    import argparse
    import sys
    from PIL import Image
    parser = argparse.ArgumentParser(
        description="Implementation of Neural Artistic Style by Gatys")
    parser.add_argument('--content', required=True)
    parser.add_argument('--style', required=True)
    parser.add_argument('--out', required=True)
    parser.add_argument('--size', type=int)
    parser.add_argument('--scale', type=float, default=1)
    parser.add_argument('--ratio', default=1, type=float)
    parser.add_argument('--device', default='cuda')
    parser.add_argument('--content-layers',
                        default=None,
                        type=lambda x: x and x.split(','))
    parser.add_argument('--iters', default=100, type=int)
    parser.add_argument('--visdom-env')
    args = parser.parse_args(sys.argv[1:])

    stylizer = NeuralStyle(device=args.device, visdom_env=args.visdom_env)

    content = Image.open(args.content)
    content.thumbnail((args.size, args.size))

    style_img = Image.open(args.style)
    if args.scale != 1.0:
        new_style_size = (int(style_img.width * args.scale),
                          int(style_img.height * args.scale))
        style_img = style_img.resize(new_style_size, Image.BICUBIC)

    result = stylizer.fit(args.iters, content, style_img, args.ratio,
                          args.content_layers)
    result = to_pil_image(result)

    result.save(args.out)



import io
import os
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import rich
import torch
from anndata import AnnData
from captum.attr import IntegratedGradients
from lightning.pytorch.callbacks import ModelCheckpoint
from scvi import REGISTRY_KEYS, settings
from scvi.data import AnnDataManager
from scvi.data.fields import (
    CategoricalObsField,
    LayerField,
    NumericalObsField,
    ObsmField,
)
from scvi.dataloaders import DataSplitter
from scvi.model.base import BaseModelClass
from scvi.train import SaveCheckpoint, TrainRunner
from scvi.utils import setup_anndata_dsp
from tqdm import tqdm

from ._data import AnnDataSplitter, _get_contact, _get_cools
from ._module import HiClipModule
from ._train import HiClipTrainingPlan
from ._utils import _logger


def _get_field(adata: AnnData, layer: str):
    if layer == "X":
        _logger.info("Using data from `adata.X`.")
        FIELD = LayerField(registry_key=layer, layer=None)
    else:
        if layer not in adata.layers:
            raise KeyError(f"{layer} is not a valid key in `adata.layers`.")
        _logger.info(f"Using data from adata.layers[{layer!r}]")
        FIELD = LayerField(
            registry_key=layer,
            layer=layer,
        )
    return FIELD


class HiClip(BaseModelClass):
    def __init__(
        self,
        adata: AnnData,
        model_name: Optional[str] = None,
        module_params: Dict[str, Any] = None,
        split_key: Optional[str] = None,
        train_split: str = "train",
        valid_split: str = "test",
        test_split: str = "ood",
    ):
        super().__init__(adata)

        self._set_attributes_maps()

        self.split_key = split_key
        self.scores = {}

        self._module = None
        self._training_plan = None
        self._data_splitter = None

        train_indices, valid_indices, test_indices = None, None, None
        if split_key is not None:
            train_indices = np.where(adata.obs.loc[:, split_key] == train_split)[0]
            valid_indices = np.where(adata.obs.loc[:, split_key] == valid_split)[0]
            test_indices = np.where(adata.obs.loc[:, split_key] == test_split)[0]

        self.train_indices = train_indices
        self.valid_indices = valid_indices
        self.test_indices = test_indices
        self.n_samples = adata.n_obs

        module_params = module_params if isinstance(module_params, Dict) else {}

        self.module = HiClipModule(
            main_loc=self.main_loc,
            sub_loc=self.sub_loc,
            target_loc=self.target_loc,
            **module_params,
        ).float()

        self._model_summary_string = self.__class__.__name__
        self._model_name = model_name
        self.init_params_ = self._get_init_params(locals())
        self.epoch_history = None

    def _set_attributes_maps(self):
        """Set attributes' maps."""
        self.main_loc = self.registry_["setup_args"]["main_layer"]
        self.sub_loc = self.registry_["setup_args"]["sub_layer"]
        self.target_loc = self.registry_["setup_args"]["target_layer"]

    @property
    def training_plan(self):
        """The model's training plan."""
        return self._training_plan

    @training_plan.setter
    def training_plan(self, plan):
        self._training_plan = plan

    @property
    def data_splitter(self):
        """Data splitter."""
        return self._data_splitter

    @data_splitter.setter
    def data_splitter(self, data_splitter):
        self._data_splitter = data_splitter

    @property
    def module(self) -> HiClipModule:
        """Model's module."""
        return self._module

    @module.setter
    def module(self, module: HiClipModule):
        self._module = module

    @property
    def model_name(self) -> str:
        """Model's name."""
        return self._model_name

    @model_name.setter
    def model_name(self, model_name: str):
        self._model_name = model_name

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        main_layer: str = "X",
        sub_layer: str = "sub",
        target_layer: str = "target",
        **kwargs: Any,
    ) -> None:
        """Setup function.

        Parameters
        ----------
        adata
            Annotated data object.
        layer
            Expression layer in :attr:`anndata.AnnData.layers` to use. If :obj:`None`, use :attr:`anndata.AnnData.X`.
        kwargs
            Keyword arguments for :meth:`~scvi.data.AnnDataManager.register_fields`.

        Returns
        -------
        Nothing, just sets up ``adata``.
        """

        MAIN_FIELD = _get_field(adata, main_layer)
        SUB_FIELD = _get_field(adata, sub_layer)
        TARGET_FIELD = _get_field(adata, target_layer)

        setup_method_args = cls._get_setup_method_args(**locals())
        adata.obs["_indices"] = np.arange(adata.n_obs)
        anndata_fields = [
            MAIN_FIELD,
            SUB_FIELD,
            TARGET_FIELD,
            NumericalObsField(REGISTRY_KEYS.INDICES_KEY, "_indices"),
        ]

        adata_manager = AnnDataManager(
            fields=anndata_fields,
            setup_method_args=setup_method_args,
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    @torch.no_grad()
    def get_dataset(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
    ) -> Dict[str, torch.Tensor]:
        """Processes :class:`~anndata.AnnData` object into valid input tensors for the model.

        Parameters
        ----------
        adata
            Annotated data object.
        indices
            Optional indices.

        Returns
        -------
        A dictionary of tensors which can be passed as input to the model.
        """
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=len(indices), shuffle=False
        )
        return list(scdl)[0]

    @torch.no_grad()
    def predict(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = 4,
    ) -> AnnData:
        """The model's prediction for a given :class:`~anndata.AnnData` object.

        Parameters
        ----------
        adata
            Annotated data object.
        indices
            Optional indices.
        batch_size
            Batch size to use.

        Returns
        -------
        Two :class:`~anndata.AnnData` objects representing the model's prediction of the expression mean and variance respectively.
        """
        self.module.eval()

        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, shuffle=False
        )
        outputs = None
        for tensors in tqdm(scdl):
            output: torch.Tensor = self.module.forward(
                tensors,
                compute_loss=False,
            )
            output = output.detach().squeeze().cpu()
            output = (
                output.unsqueeze(0) if output.dim() != 3 else output
            )  # [batch_size, n_bin, n_bin]
            output = np.array([mat.flatten() for mat in output.numpy()])
            outputs = output if outputs is None else np.concatenate((outputs, output))

        _adata = adata[indices] if indices is not None else adata
        pred_adata = AnnData(X=outputs, obs=_adata.obs.copy())
        pred_adata.obs_names = _adata.obs_names
        pred_adata.var_names = _adata.var_names

        return pred_adata

    @torch.no_grad()
    def observe(
        self,
        main_cooler_uri: str,
        sub_cooler_uri: str,
        chrom: str,
        start: int,
        end: int,
    ) -> np.array:
        self.module.eval()

        main_c, sub_c = _get_cools(main_cooler_uri, sub_cooler_uri)
        assert main_c.binsize == sub_c.binsize

        _main = _get_contact(main_c, chrom, start, end)
        _sub = _get_contact(sub_c, chrom, start, end)
        tensors = dict()
        tensors[self.main_loc] = torch.Tensor(_main.flatten()[np.newaxis, :])
        tensors[self.sub_loc] = torch.Tensor(_sub.flatten()[np.newaxis, :])

        output: torch.Tensor = self.module.forward(
            tensors,
            compute_loss=False,
        )
        output = output.detach().squeeze().cpu()  # [n_bin, n_bin]

        return np.array(output)

    @torch.no_grad()
    def interpret(
        self,
        main_cooler_uri: str,
        sub_cooler_uri: str,
        chrom: str,
        start: int,
        end: int,
        target_index: List[Tuple[int, int]] = None,
        internal_batch_size: int = 8,
    ):
        self.module.eval()
        ig = IntegratedGradients(self.module.inference)

        main_c, sub_c = _get_cools(main_cooler_uri, sub_cooler_uri)
        assert main_c.binsize == sub_c.binsize

        _main = _get_contact(main_c, chrom, start, end)
        _sub = _get_contact(sub_c, chrom, start, end)
        tensors = dict()
        tensors[self.main_loc] = torch.Tensor(_main.flatten()[np.newaxis, :])
        tensors[self.sub_loc] = torch.Tensor(_sub.flatten()[np.newaxis, :])

        attributions = dict()
        if target_index is None:
            target_index = [(i, j) for i in range(_main.shape[-2]) for j in range(_main.shape[-1])]
        for i, j in tqdm(target_index):
            inference_inputs = self.module._get_inference_input(tensors)
            _attributions = ig.attribute(
                (inference_inputs["main"], inference_inputs["sub"]),
                target=(0, i, j),
                internal_batch_size=internal_batch_size,
            )
            attributions[i, j] = {
                "main": _attributions[0],
                "sub": _attributions[1],
            }
        return attributions

    def save(
        self,
        dir_path: Optional[str] = None,
        overwrite: bool = False,
        save_anndata: bool = False,
        **anndata_save_kwargs: Any,
    ) -> None:
        """Save the model.

        Parameters
        ----------
        dir_path
            Directory where to save the model. If :obj:`None`, it will be determined automatically.
        overwrite
            Whether to overwrite an existing model.
        save_anndata
            Whether to also save :class:`~anndata.AnnData`.
        anndata_save_kwargs
            Keyword arguments :meth:`scvi.model.base.BaseModelClass.save`.

        Returns
        -------
        Nothing, just saves the model.
        """
        if dir_path is None:
            dir_path = (
                f"./{self.__class__.__name__}_model/"
                if self.model_name is None
                else f"./{self.__class__.__name__}_{self.model_name}_model/"
            )
        super().save(
            dir_path=dir_path,
            overwrite=overwrite,
            save_anndata=save_anndata,
            **anndata_save_kwargs,
        )

        if isinstance(self.training_plan.epoch_history, dict):
            self.epoch_history = pd.DataFrame().from_dict(
                self.training_plan.epoch_history
            )
            self.epoch_history.to_csv(
                os.path.join(dir_path, "history.csv"), index=False
            )

    @classmethod
    def load(
        cls,
        dir_path: str,
        adata: Optional[AnnData] = None,
        accelerator: str = "auto",
        **kwargs: Any,
    ) -> "HiClip":
        """Load a saved model.

        Parameters
        ----------
        dir_path
            Directory where the model is saved.
        adata
            AnnData organized in the same way as data used to train model.
        accelerator
            gpu or cpu.
        kwargs
            Keyword arguments for :meth:`scvi`

        Returns
        -------
        The saved model.
        """
        model = super().load(dir_path, adata, accelerator=accelerator, **kwargs)

        fname = os.path.join(dir_path, "history.csv")
        if os.path.isfile(fname):
            model.epoch_history = pd.read_csv(fname)
        else:
            _logger.warning(f"The history file `{fname}` was not found")

        return model

    def train(
        self,
        max_epochs: Optional[int] = None,
        accelerator: str = "auto",
        train_size: float = 0.9,
        validation_size: Optional[float] = None,
        plan_kwargs: Optional[Dict[str, Any]] = None,
        batch_size: int = 4,
        early_stopping: bool = False,
        **trainer_kwargs: Any,
    ) -> None:
        """Train the :class:`~hiclip.HiClip` model.

        Parameters
        ----------
        max_epochs
            Maximum number of epochs for training.
        accelerator
            gpu or cpu.
        train_size
            Fraction of training data in the case of randomly splitting dataset to train/validation
            if :attr:`split_key` is not set in model's constructor.
        validation_size
            Fraction of validation data in the case of randomly splitting dataset to train/validation
            if :attr:`split_key` is not set in model's constructor.
        batch_size
            Size of mini-batches for training.
        early_stopping
            If `True`, early stopping will be used during training on validation dataset.
        plan_kwargs
            Keyword arguments for :class:`~scvi.train.TrainingPlan`.
        trainer_kwargs
            Keyword arguments for :class:`~scvi.train.TrainRunner`.

        Returns
        -------
        Nothing, just trains the :class:`~hiclip.HiClip` model.
        """
        plan_kwargs = plan_kwargs if plan_kwargs is not None else {}
        self.training_plan = HiClipTrainingPlan(
            module=self.module,
            **plan_kwargs,
        )

        monitor = trainer_kwargs.pop("monitor", "val_hiclip_metric")
        save_ckpt_every_n_epoch = trainer_kwargs.pop("save_ckpt_every_n_epoch", 20)
        enable_checkpointing = trainer_kwargs.pop("enable_checkpointing", True)

        trainer_kwargs["callbacks"] = (
            []
            if "callbacks" not in trainer_kwargs.keys()
            else trainer_kwargs["callbacks"]
        )
        if enable_checkpointing:
            checkpointing_callback = SaveCheckpoint(
                monitor=monitor, every_n_epochs=save_ckpt_every_n_epoch
            )
            trainer_kwargs["callbacks"] += [checkpointing_callback]
            trainer_kwargs["checkpointing_monitor"] = monitor

        num_workers = trainer_kwargs.pop("num_workers", 0)
        if max_epochs is None:
            n_cells = self.adata.n_obs
            max_epochs = np.min([round((20000 / n_cells) * 400), 400])

        manual_splitting = (
            (self.valid_indices is not None)
            and (self.train_indices is not None)
            and (self.test_indices is not None)
        )
        if manual_splitting:
            self.data_splitter = AnnDataSplitter(
                self.adata_manager,
                train_indices=self.train_indices,
                valid_indices=self.valid_indices,
                test_indices=self.test_indices,
                batch_size=batch_size,
                accelerator=accelerator,
                num_workers=num_workers,
            )
        else:
            self.data_splitter = DataSplitter(
                self.adata_manager,
                train_size=train_size,
                validation_size=validation_size,
                batch_size=batch_size,
                num_workers=num_workers,
            )
        es = "early_stopping"
        trainer_kwargs[es] = (
            early_stopping if es not in trainer_kwargs.keys() else trainer_kwargs[es]
        )

        trainer_kwargs["check_val_every_n_epoch"] = trainer_kwargs.get(
            "check_val_every_n_epoch", 1
        )
        trainer_kwargs["early_stopping_patience"] = trainer_kwargs.get(
            "early_stopping_patience", 20
        )

        root_dir = settings.logging_dir
        root_dir = (
            os.path.join(root_dir, f"{self.__class__.__name__}/")
            if self.model_name is None
            else os.path.join(root_dir, f"{self.model_name}_{self.__class__.__name__}/")
        )

        trainer_kwargs["default_root_dir"] = trainer_kwargs.pop(
            "default_root_dir", root_dir
        )

        runner = TrainRunner(
            self,
            training_plan=self.training_plan,
            data_splitter=self.data_splitter,
            max_epochs=max_epochs,
            accelerator=accelerator,
            early_stopping_monitor=monitor,
            early_stopping_mode="max",
            enable_checkpointing=enable_checkpointing,
            **trainer_kwargs,
        )

        return runner()

    def __repr__(self) -> str:
        buffer = io.StringIO()
        summary_string = f"{self._model_summary_string} training status: "
        summary_string += "{}".format(
            "[green]Trained[/]" if self.is_trained else "[red]Not trained[/]"
        )
        console = rich.console.Console(file=buffer)
        with console.capture() as capture:
            console.print(summary_string)
        return capture.get()



import torch
import torch.nn as nn
from .conv import Conv2d, Conv3x3, Conv1x1
from torchelie.utils import kaiming, constant_init
from .condseq import CondSeq
from .interpolate import InterpolateBilinear2d
import collections
from typing import List, Tuple, Optional, cast
from .utils import remove_batchnorm, insert_before


class SEBlock(nn.Module):
    """
    A Squeeze-And-Excite block

    Args:
        in_ch (int): input channels
        reduction (int): channels reduction factor for the hidden number of
            channels
    """
    def __init__(self, in_ch: int, reduction: int = 16) -> None:
        super(SEBlock, self).__init__()
        reduc = in_ch // reduction
        self.proj = nn.Sequential(
            collections.OrderedDict([('pool', nn.AdaptiveAvgPool2d(1)),
                                     ('squeeze', kaiming(Conv1x1(in_ch,
                                                                 reduc))),
                                     ('relu', nn.ReLU(True)),
                                     ('excite', kaiming(Conv1x1(reduc,
                                                                in_ch))),
                                     ('attn', nn.Sigmoid())]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.proj(x)


def _make_resnet_shortcut(in_channels: int, out_channels: int,
                          stride: int) -> CondSeq:
    shortcut = CondSeq()
    if stride != 1:
        shortcut.add_module(
            'pool', nn.AvgPool2d(1 + (stride - 1) * 2, stride, stride // 2))

    if in_channels != out_channels:
        shortcut.add_module(
            'conv', kaiming(Conv1x1(in_channels, out_channels, bias=False)))
        shortcut.add_module('bn', nn.BatchNorm2d(out_channels))
    return shortcut


class ResBlockBottleneck(nn.Module):
    """
    A Residual Block. Skip connection will be added if the number of input and
    output channels don't match or stride > 1 is used.

    Args:
        in_ch (int): input channels
        out_ch (int): output channels
        stride (int): stride
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 stride: int = 1) -> None:
        super(ResBlockBottleneck, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

        self.pre = CondSeq()

        self.wide(divisor=4)

        self.relu = nn.ReLU(True)

        self.shortcut = _make_resnet_shortcut(in_channels, out_channels, stride)

        self.post = CondSeq()

    def condition(self, z: torch.Tensor) -> None:
        self.branch.condition(z)
        self.shortcut.condition(z)
        self.pre.condition(z)
        self.post.condition(z)

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        if z is not None:
            self.condition(z)

        x = self.pre(x)
        x = self.relu(self.branch(x).add_(self.shortcut(x)))
        return self.post(x)

    def remove_batchnorm(self) -> 'ResBlockBottleneck':
        remove_batchnorm(self.branch)
        remove_batchnorm(self.shortcut)
        assert isinstance(self.branch.conv3, nn.Conv2d)
        constant_init(self.branch.conv3, 0)

        return self

    def use_se(self) -> 'ResBlockBottleneck':
        self.branch.add_module('se', SEBlock(self.out_channels))
        return self

    def resnext(self) -> 'ResBlockBottleneck':
        self.wide(divisor=2)
        c = self.branch.conv2
        assert isinstance(c, nn.Conv2d)
        self.branch.conv2 = kaiming(
            nn.Conv2d(c.in_channels,
                      c.out_channels,
                      kernel_size=3,
                      padding=1,
                      stride=self.stride,
                      bias=c.bias is not None,
                      groups=32))
        return self

    def wide(self, divisor: int = 2) -> 'ResBlockBottleneck':
        in_ch = self.in_channels
        out_ch = self.out_channels
        mid = out_ch // divisor
        stride = self.stride

        self.branch = CondSeq(
            collections.OrderedDict([
                ('conv1', kaiming(Conv1x1(in_ch, mid, bias=False))),
                ('bn1', constant_init(nn.BatchNorm2d(mid), 1)),
                ('relu', nn.ReLU(True)),
                ('conv2', kaiming(Conv3x3(mid, mid, stride=stride,
                                          bias=False))),
                ('bn2', constant_init(nn.BatchNorm2d(mid), 1)),
                ('relu2', nn.ReLU(True)),
                ('conv3', kaiming(Conv1x1(mid, out_ch, bias=False))),
                ('bn3', constant_init(nn.BatchNorm2d(out_ch), 0)),
            ]))
        return self

    def upsample_instead(self) -> 'ResBlockBottleneck':
        if self.stride == 1:
            return self

        assert isinstance(self.branch.conv2, nn.Conv2d)
        self.branch.conv2.stride = (1, 1)
        insert_before(self.branch, 'conv2',
                      InterpolateBilinear2d(scale_factor=self.stride),
                      'upsample')

        if hasattr(self.shortcut, 'pool'):
            self.shortcut.pool = InterpolateBilinear2d(
                scale_factor=self.stride)
        return self


class ResBlock(nn.Module):
    """
    A Residual Block. Skip connection will be added if the number of input and
    output channels don't match or stride > 1 is used.

    Args:
        in_channels (int): input channels
        out_channels (int): output channels
        stride (int): stride
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 stride: int = 1) -> None:
        super(ResBlock, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

        self.pre = CondSeq()

        self.branch = CondSeq(
            collections.OrderedDict([
                ('conv1',
                 kaiming(
                     Conv3x3(in_channels,
                             out_channels,
                             stride=stride,
                             bias=False))),
                ('bn1', constant_init(nn.BatchNorm2d(out_channels), 1)),
                ('relu', nn.ReLU(True)),
                ('conv2',
                 kaiming(Conv3x3(out_channels, out_channels, bias=False))),
                ('bn2', constant_init(nn.BatchNorm2d(out_channels), 0))
            ]))

        self.shortcut = _make_resnet_shortcut(in_channels, out_channels, stride)

        self.relu = nn.ReLU(True)

        self.post = CondSeq()

    def condition(self, z: torch.Tensor) -> None:
        self.branch.condition(z)
        self.shortcut.condition(z)
        self.pre.condition(z)
        self.post.condition(z)

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        if z is not None:
            self.condition(z)

        x = self.pre(x)
        x = self.relu(self.branch(x).add_(self.shortcut(x)))
        return self.post(x)

    def remove_batchnorm(self) -> 'ResBlock':
        remove_batchnorm(self.branch)
        remove_batchnorm(self.shortcut)
        assert isinstance(self.branch.conv2, nn.Conv2d)
        constant_init(self.branch.conv2, 0)
        return self

    def use_se(self) -> 'ResBlock':
        self.branch.add_module('se', SEBlock(self.out_channels))
        return self


def make_preact_resnet_shortcut(in_ch: int, out_ch: int,
                                stride: int) -> CondSeq:
    sc: List[Tuple[str, nn.Module]] = []
    if stride != 1:
        sc.append(
            ('pool', nn.AvgPool2d(1 + (stride - 1) * 2, stride, stride // 2)))

    if in_ch != out_ch:
        sc.append(('conv', kaiming(Conv1x1(in_ch, out_ch))))

    return CondSeq(collections.OrderedDict(sc))


class PreactResBlock(nn.Module):
    """
    A Preactivated Residual Block. Skip connection will be added if the number
    of input and output channels don't match or stride > 1 is used.

    Args:
        in_ch (int): input channels
        out_ch (int): output channels
        stride (int): stride
    """
    branch: CondSeq
    shortcut: CondSeq

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 stride: int = 1) -> None:
        super(PreactResBlock, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

        self.pre = CondSeq()
        self.preact = CondSeq()

        self.branch = CondSeq(
            collections.OrderedDict([
                ('bn1', constant_init(nn.BatchNorm2d(in_channels), 1)),
                ('relu', nn.ReLU(True)),
                ('conv1',
                 kaiming(
                     Conv3x3(in_channels,
                             out_channels,
                             stride=stride,
                             bias=False))),
                ('bn2', constant_init(nn.BatchNorm2d(out_channels), 1)),
                ('relu2', nn.ReLU(True)),
                ('conv2', constant_init(Conv3x3(out_channels, out_channels),
                                        0)),
            ]))

        self.shortcut = make_preact_resnet_shortcut(in_channels, out_channels,
                                                    stride)

        if in_channels != out_channels:
            self.preact_skip()

        self.post = CondSeq()

    def remove_batchnorm(self) -> 'PreactResBlock':
        remove_batchnorm(self.branch)
        remove_batchnorm(self.shortcut)
        remove_batchnorm(self.preact)
        return self

    def condition(self, z: torch.Tensor) -> None:
        self.pre.condition(z)
        self.preact.condition(z)
        self.branch.condition(z)
        self.shortcut.condition(z)
        self.post.condition(z)

    def preact_skip(self) -> 'PreactResBlock':
        if hasattr(self.branch, 'bn1'):
            self.preact.add_module('bn1', cast(nn.Module, self.branch.bn1))
            del self.branch.bn1

        self.preact.add_module('relu', cast(nn.Module, self.branch.relu))
        del self.branch.relu

        return self

    def no_preact(self) -> 'PreactResBlock':
        self.preact_skip()
        self.preact = CondSeq()
        return self

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        if z is not None:
            self.condition(z)
        x = self.pre(x)
        x = self.preact(x)
        x = self.shortcut(x) + self.branch(x)
        return self.post(x)

    def use_se(self) -> 'PreactResBlock':
        self.branch.add_module('se', SEBlock(self.out_channels))
        return self


class PreactResBlockBottleneck(nn.Module):
    """
    A Preactivated Residual Block. Skip connection will be added if the number
    of input and output channels don't match or stride > 1 is used.

    Args:
        in_ch (int): input channels
        out_ch (int): output channels
        stride (int): stride
    """
    branch: CondSeq
    shortcut: CondSeq

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 stride: int = 1) -> None:
        super(PreactResBlockBottleneck, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

        self.pre = CondSeq()
        self.preact = CondSeq()

        self.branch = CondSeq()
        self.wide(divisor=4)

        self.shortcut = make_preact_resnet_shortcut(in_channels, out_channels,
                                                    stride)

        if in_channels != out_channels:
            self.preact_skip()

        self.post = CondSeq()

    def remove_batchnorm(self) -> 'PreactResBlockBottleneck':
        remove_batchnorm(self.branch)
        remove_batchnorm(self.shortcut)
        remove_batchnorm(self.preact)
        return self

    def condition(self, z: torch.Tensor) -> None:
        self.pre.condition(z)
        self.preact.condition(z)
        self.branch.condition(z)
        self.shortcut.condition(z)
        self.post.condition(z)

    def preact_skip(self) -> 'PreactResBlockBottleneck':
        if hasattr(self.branch, 'bn1'):
            self.preact.add_module('bn1', cast(nn.Module, self.branch.bn1))
            del self.branch.bn1

        self.preact.add_module('relu', cast(nn.Module, self.branch.relu))
        del self.branch.relu

        return self

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        if z is not None:
            self.condition(z)
        x = self.pre(x)
        x = self.preact(x)
        x = self.shortcut(x) + self.branch(x)
        return self.post(x)

    def use_se(self) -> 'PreactResBlockBottleneck':
        self.branch.add_module('se', SEBlock(self.out_channels))
        return self

    def no_preact(self) -> 'PreactResBlockBottleneck':
        self.preact_skip()
        self.preact = CondSeq()
        return self

    def resnext(self, groups: int = 32) -> 'PreactResBlockBottleneck':
        self.wide(divisor=4)
        c = self.branch.conv2
        assert isinstance(c, nn.Conv2d)
        self.branch.conv2 = kaiming(
            nn.Conv2d(c.in_channels,
                      c.out_channels,
                      3,
                      groups=groups,
                      stride=self.stride,
                      padding=1,
                      bias=False))
        return self

    def wide(self, divisor: int = 2) -> 'PreactResBlockBottleneck':
        in_ch = self.in_channels
        out_ch = self.out_channels
        stride = self.stride
        mid = out_ch // divisor
        self.branch = CondSeq(
            collections.OrderedDict([
                ('bn1', constant_init(nn.BatchNorm2d(in_ch), 1)),
                ('relu', nn.ReLU(True)),
                ('conv1', kaiming(Conv1x1(in_ch, mid, bias=False))),
                ('bn2', constant_init(nn.BatchNorm2d(mid), 1)),
                ('relu2', nn.ReLU(True)),
                ('conv2', kaiming(Conv3x3(mid, mid, stride=stride,
                                          bias=False))),
                ('bn3', constant_init(nn.BatchNorm2d(mid), 1)),
                ('relu3', nn.ReLU(True)),
                ('conv3', constant_init(Conv1x1(mid, out_ch), 0))
            ]))
        return self



"""
Those classes are different ways of averaging metrics.
"""

import torchelie.utils as tu
from typing import List, Optional


class RunningAvg(tu.AutoStateDict):
    """
    Average by keeping the whole sum and number of elements of the data logged.
    Useful when the metrics come per batch and an accurate number for the whole
    epoch is needed.
    """
    def __init__(self) -> None:
        super(RunningAvg, self).__init__()
        self.count = 0.0
        self.val = 0.0

    def log(self, x: float, total: int = 1):
        """
        Log metric

        Args:
            x: the metric
            total: how many element does this represent
        """
        self.count += total
        self.val += x

    def get(self) -> float:
        """
        Get the average so far
        """
        if self.count == 0:
            return float('nan')
        return self.val / self.count


class WindowAvg(tu.AutoStateDict):
    """
    Average a window containing the `k` previous logged values

    Args:
        k (int): the window's length
    """
    def __init__(self, k: int = 100) -> None:
        super(WindowAvg, self).__init__()
        self.vals: List[float] = []
        self.k = k

    def log(self, x: float) -> None:
        """
        Log `x`
        """
        if len(self.vals) == self.k:
            self.vals = self.vals[1:]
        self.vals.append(x)

    def get(self) -> float:
        """
        Return the value averaged over the window
        """
        if len(self.vals) == 0:
            return float("nan")
        return sum(self.vals) / len(self.vals)


class ExponentialAvg(tu.AutoStateDict):
    r"""
    Keep an exponentially decaying average of the values according to

    :math:`y := \beta y + (1 - \beta) x`

    Args:
        beta (float): the decay rate
    """
    def __init__(self, beta: float = 0.6):
        super(ExponentialAvg, self).__init__()
        self.beta = beta
        self.val: Optional[float] = None

    def log(self, x: float) -> None:
        """Log `x`"""
        if self.val is None:
            self.val = x
        else:
            self.val = self.beta * self.val + (1 - self.beta) * x

    def get(self) -> float:
        """Return the exponential average at this time step"""
        assert self.val is not None, 'no value yet'
        return self.val



import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


class L2Constraint(nn.Module):
    """
    From `Ranjan 2017`_ , L2-constrained Softmax Loss for Discriminative Face
    Verification.

    Args:
        dim (int): number of channels of the feature vector
        num_classes (int): number of identities
        fixed (bool): whether to use the fixed or dynamic version of AdaCos
            (default: False)

    :: _Ranjan 2017: https://arxiv.org/abs/1703.09507
    """
    def __init__(self, dim: int, num_classes: int, s: float = 30.):
        super(L2Constraint, self).__init__()
        self.weight = nn.Parameter(torch.FloatTensor(num_classes, dim))
        nn.init.orthogonal_(self.weight)
        self.num_classes = num_classes
        self.s = s

    def forward(self, input: torch.Tensor,
                label: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass

        Args:
            input (tensor): feature vectors
            label (tensor): labels

        Returns:
            scaled cosine logits, cosine logits
        """
        cosine = F.linear(F.normalize(input), F.normalize(self.weight))
        output = cosine * self.s

        return output, cosine

    def __repr__(self):
        return "L2CrossEntropy(s={})".format(self.s)


class AdaCos(nn.Module):
    """
    From AdaCos: [Adaptively Scaling Cosine Logits for Effectively Learning
    Deep Face Representations](https://arxiv.org/abs/1905.00292)

    Args:
        dim (int): number of channels of the feature vector
        num_classes (int): number of identities
        fixed (bool): whether to use the fixed or dynamic version of AdaCos
            (default: False)
        estimate_B (bool): is using dynamic AdaCos, B is estimated from the
            real angles of the cosine similarity. However I found that this
            method was not numerically stable and experimented with the
            approximation :code:`B = num_classes - 1` that was more satisfying.
    """
    s: torch.Tensor

    def __init__(self,
                 dim: int,
                 num_classes: int,
                 fixed: bool = False,
                 estimate_B: bool = False):
        super(AdaCos, self).__init__()
        self.fixed = fixed
        self.weight = nn.Parameter(torch.FloatTensor(num_classes, dim))
        nn.init.xavier_normal_(self.weight)
        self.num_classes = num_classes
        self.register_buffer(
            's', torch.tensor(math.sqrt(2) * math.log(num_classes - 1)))
        self.register_buffer('B', torch.tensor(num_classes - 1.))
        self.estimate_B = estimate_B

    def forward(
        self, input: torch.Tensor, label: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass

        Args:
            input (tensor): feature vectors
            label (tensor): labels

        Returns:
            scaled cosine logits, cosine logits
        """
        cosine = F.linear(F.normalize(input), F.normalize(self.weight))

        if self.fixed:
            return cosine * self.s, cosine

        if self.training:
            with torch.no_grad():
                correct_cos = torch.gather(cosine, 1, label.unsqueeze(1))
                theta_med = torch.acos(correct_cos).median()

                if self.estimate_B:
                    output = cosine * self.s
                    expout = output.exp()
                    correct_expout = torch.gather(expout, 1,
                                                  label.unsqueeze(1))
                    correct_expout.squeeze_()
                    self.B = torch.mean(expout.sum(1) - correct_expout, dim=0)

                self.s = torch.log(self.B) / math.cos(
                    min(math.pi / 4, theta_med.item()))
        return cosine * self.s, cosine

    def __repr__(self) -> str:
        return "FixedAdaCos(fixed={})".format(self.fixed)



import torch
import torch.nn.functional as F
from torch.autograd import Function
from .vq import quantize


def laplacian(images, n_down=4):
    """
    Decompose a 4D images tensor into a laplacian pyramid.

    Args:
        images (4D Tensor): source images
        n_down (int): how many times to downscale the input

    Returns:
        A list of tensor: laplacian pyramid
    """
    lapls = []

    for i in range(n_down):
        n = F.interpolate(images,
                          scale_factor=0.5,
                          mode='bilinear',
                          align_corners=True)
        lapls.append(images - F.interpolate(
            n, size=images.shape[-2:], mode='bilinear', align_corners=True))
        images = n

    lapls.append(images)
    return lapls


def combine_laplacians(laplacians):
    """
    Recombine a list of of tensor as returned by :code:`laplacian()` into an
    image batch

    Args:
        laplacians (list of tensors: laplacian pyramid

    Returns:
        a tensor
    """
    biggest = laplacians[0]

    rescaled = [biggest]
    for im in laplacians[1:]:
        rescaled.append(
            F.interpolate(im,
                          size=biggest.shape[-2:],
                          mode='bilinear',
                          align_corners=True))

    mixed = torch.stack(rescaled, dim=-1)
    return mixed.sum(dim=-1)


class InformationBottleneckFunc(Function):
    @staticmethod
    def forward(ctx, mu, sigma, strength=1):
        z = torch.randn_like(mu)
        x = mu + z * sigma
        s = torch.tensor(strength, device=x.device)
        ctx.save_for_backward(mu, sigma, z, s)
        return x

    @staticmethod
    def backward(ctx, d_out):
        mu, sigma, z, strength = ctx.saved_tensors

        # kl = -0.5 * (1 + log(sigma^2) - mu^2 - sigma^2)
        # dkl/dmu = -0.5 * (-2*mu)
        #         = mu
        # dkl/dsig = -0.5 * (d 2*log(sigma) - d sigma^2)
        #          = -0.5 * (2/sigma - 2*sigma)
        #          = -1/sigma + sigma
        d_mu = d_out + strength * mu
        d_sigma = z * d_out
        d_sigma += -strength / sigma + strength * sigma
        return d_mu, d_sigma, None


information_bottleneck = InformationBottleneckFunc.apply
unit_gaussian_prior = InformationBottleneckFunc.apply



import random
import multiprocessing

import torchelie.utils as tu
from torchelie.datasets.debug import *
from .concat import HorizontalConcatDataset, MergedDataset
from .ms1m import MS1M
from .pix2pix import Pix2PixDataset, ImagesPaths, UnlabeledImages
from .pix2pix import SideBySideImagePairsDataset

from typing import Sequence
from torchvision.datasets import ImageFolder
from pathlib import Path
import os
import torch


def FastImageFolder(root: str, *args, **kwargs) -> ImageFolder:
    """
    Load an ImageFolder dataset faster by caching the file list the first
    time it is accessed. Force refreshing by deleting
    :code:`{root}/cached_list.pth`.

    Forwards all arguments to torchvision.datasets.ImageFolder

    Returns:
        The ImageFolder transparently loaded faster.
    """
    cache = Path(root) / 'cached_list.pth'
    fields = ['samples', 'imgs', 'classes', 'class_to_idx']
    if cache.exists():
        os.makedirs('/tmp/empty/0', exist_ok=True)
        open('/tmp/empty/0/0.jpg', 'a').close()
        ds = ImageFolder('/tmp/empty', *args, **kwargs)
        save = torch.load(str(cache))
        for f in fields:
            setattr(ds, f, save[f])
    else:
        ds = ImageFolder(root, *args, **kwargs)
        torch.save({f: getattr(ds, f) for f in fields}, str(cache))
    return ds


class PairedDataset(torch.utils.data.Dataset):
    """
    A dataset that returns all possible pairs of samples of two datasets

    Args:
        dataset1 (Dataset): a dataset
        dataset2 (Dataset): another dataset
    """
    def __init__(self, dataset1: Sequence, dataset2: Sequence) -> None:
        super(PairedDataset, self).__init__()
        self.dataset1 = dataset1
        self.dataset2 = dataset2

    def __getitem__(self, i):
        idx1 = i % len(self.dataset1)
        idx2 = i // len(self.dataset2)

        x1 = self.dataset1[idx1]
        x2 = self.dataset2[idx2]

        if isinstance(x1, (tuple, list)):
            return list(zip(x1, x2))
        return x1, x2

    def __len__(self) -> int:
        return len(self.dataset1) * len(self.dataset2)

    def __repr__(self) -> str:
        return ("PairedDataset:\n" + tu.indent(repr(self.dataset1))
                + "\n--\n" + tu.indent(repr(self.dataset2)))


class RandomPairsDataset(torch.utils.data.Dataset):
    """
    :code:`PairedDataset` deterministically constructs all possible pairs of
    items from both datasets. This may cause some problems while sampling
    if permutation or weight array must be constructed as they would be of size
    :code:`len(dataset1) * len(dataset2)` and may not fit in memory.

    :code:`RandomPairsDataset` presents itself as dataset of 10000 elements,
    and accessing the ith item returns a random pair, effectively ignoring i.

    Args:
        dataset1 (Dataset): a dataset
        dataset2 (Dataset): another dataset
    """
    def __init__(self, dataset1, dataset2):
        super(RandomPairsDataset, self).__init__()
        self.dataset1 = dataset1
        self.dataset2 = dataset2

    def __getitem__(self, i):
        idx1 = random.randrange(0, len(self.dataset1) - 1)
        idx2 = random.randrange(0, len(self.dataset2) - 1)

        x1 = self.dataset1[idx1]
        x2 = self.dataset2[idx2]

        if isinstance(x1, (tuple, list)):
            return list(zip(x1, x2))
        return x1, x2

    def __len__(self):
        return 10000

    def __repr__(self):
        return ("PairedDataset:\n" + tu.indent(repr(self.dataset1))
                + "\n--\n" + tu.indent(repr(self.dataset2)))


def mixup(x1, x2, y1, y2, num_classes, mixer=None, alpha=0.4):
    r"""
    Mixes samples `x1` and `x2` with respective labels `y1` and `y2` according
    to MixUp

    :math:`\lambda \sim \text{Beta}(\alpha, \alpha)`

    :math:`x = \lambda x_1 + (1-\lambda) x_2`

    :math:`y = \lambda y_1 + (1 - \lambda) y_2`

    Args:
        x1 (tensor): sample 1
        x2 (tensor): sample 2
        y1 (tensor): label 1
        y2 (tensor): label 2
        num_classes (int): number of classes
        mixer (Distribution, optional): a distribution to sample lambda from.
            If unspecified, the distribution will be a Beta(alpha, alpha)
        alpha (float): if mixer is unspecified, used to parameterize the Beta
            distribution
    """
    if mixer is None:
        alpha = torch.tensor([alpha])
        mixer = torch.distributions.Beta(alpha, alpha)

    y1 = torch.tensor(y1)
    y2 = torch.tensor(y2)
    lam = mixer.sample(y1.shape).to(y1.device)
    y1 = torch.nn.functional.one_hot(y1, num_classes=num_classes).float().to(
        y1.device)
    y2 = torch.nn.functional.one_hot(y2, num_classes=num_classes).float().to(
        y1.device)

    return (lam * x1 + (1 - lam) * x2), (lam * y1 + (1 - lam) * y2)


class _Wrap:
    def __init__(self, instance):
        # FIXME: NOT WORKING WHEN SETTING MEMBERS
        # self.__dict__ = instance.__dict__
        self.ds = instance

    def __getattr__(self, attr):
        return getattr(self.ds, attr)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)


class MixUpDataset(_Wrap):
    """
    Linearly mixes two samples and labels from a dataset according to the MixUp
    algorithm

    https://arxiv.org/abs/1905.02249

    Args:
        dataset (Dataset): the dataset
        alpha (float): the alpha that parameterizes the beta distribution from
            which the blending factor is sampled
    """
    def __init__(self, dataset, alpha=0.4):
        super(MixUpDataset, self).__init__(dataset)
        self.ds = dataset
        alpha = torch.tensor([alpha])
        self.mixer = torch.distributions.Beta(alpha, alpha)

    def __len__(self):
        return len(self.ds)

    def __getitem__(self, i):
        x1, y1 = self.ds[i]
        x2, y2 = random.choice(self.ds)

        return mixup(x1, x2, y1, y2, len(self.ds.classes), self.mixer)

    def __repr__(self):
        return f"MixUpDataset({self.mixer}):\n" + tu.indent(repr(self.ds))


class _Proxy:
    def __init__(self, ds, indices, remap_unused_classes, cls_map):
        self.ds = ds
        self.indices = indices
        self.remap_unused_classes = remap_unused_classes
        self.cls_map = cls_map

    def __len__(self):
        return len(indices)

    def __getitem__(self, i):
        b = self.ds.samples[self.indices[i]]
        if self.remap_unused_classes:
            b = list(b)
            b[1] = self.cls_map[b[1]]
        return b


class Subset:
    """
    Create a subset that is a random ratio of a dataset.

    Args:
        ds (Dataset): the dataset to sample from. Must have a :code:`.samples`
            member like torchvision's datasets.
        ratio (float): a value between 0 and 1, the subsampling ratio.
        remap_unused_classes (boolean): if True, classes not represented in the
            subset will not be considered. Remaining classes will be numbered
            from 0 to N.
    """
    def __init__(self, ds, ratio, remap_unused_classes=False):
        self.ratio = ratio
        self.ds = ds
        indices = [i for i in range(len(ds)) if random.uniform(0, 1) < ratio]
        self.indices = indices

        self.remap_classes = remap_unused_classes
        if remap_unused_classes:
            cls_map = {}
            cls = []
            cls_to_idx = {}
            for i in self.indices:
                c = ds.samples[i][1]
                if c not in cls_map:
                    new_idx = len(cls_map)
                    cls_map[c] = new_idx
                    cls.append(ds.classes[c])
                    cls_to_idx[ds.classes[c]] = new_idx
            self.cls_map = cls_map
            self.classes = cls
            self.class_to_idx = cls_to_idx
        else:
            cls_map = {}
            self.classes = ds.classes
            self.class_to_idx = ds.class_to_idx

        self.samples = _Proxy(ds, indices, remap_unused_classes, cls_map)
        self.imgs = self.samples

    def __repr__(self):
        return "Subset(len={}, n_classes={}):\n{}".format(
            len(self.indices), len(self.classes), tu.indent(repr(self.ds)))

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, i):
        b = self.ds[self.indices[i]]
        if self.remap_classes:
            b = list(b)
            b[1] = self.cls_map[b[1]]
        return b


class NoexceptDataset(_Wrap):
    """
    Wrap a dataset and absorbs the exceptions it raises.  Useful in case of a
    big downloaded dataset with corrupted samples for instance.

    Args:
        ds (Dataset): a dataset
    """
    def __init__(self, ds):
        super(NoexceptDataset, self).__init__(ds)
        self.ds = ds

    def __len__(self):
        return len(self.ds)

    def __getitem__(self, i):
        while True:
            try:
                return self.ds[i]
            except Exception as e:
                print(e)
            if i < len(self) - 1:
                i += 1
            else:
                i = 0

    def __repr__(self):
        return "NoexceptDataset:\n{}".format(tu.indent(repr(self.ds)))


class WithIndexDataset(_Wrap):
    """
    Wrap a dataset. Also returns the index of the accessed element. Original
    dataset's attributes are transparently accessible

    Args:
        ds (Dataset): A dataset
    """
    def __init__(self, ds):
        super(WithIndexDataset, self).__init__(ds)
        self.ds = ds

    def __getitem__(self, i):
        """
        Args:
            i (int): index

        Returns:
            A tuple (i, self.ds[i])
        """
        return i, self.ds[i]

    def __len__(self):
        return len(self.ds)

    def __repr__(self):
        return "WithIndexDataset:\n{}".format(tu.indent(repr(self.ds)))


class CachedDataset(_Wrap):
    """
    Wrap a dataset. Lazily caches *in RAM* elements returned by the underlying
    dataset.

    Args:
        ds (Dataset): A dataset
        transform (Callable): transform to apply on cached elements
        device: the device on which the cache is allocated
    """
    def __init__(self, ds, transform=None, device='cpu'):
        super(CachedDataset, self).__init__(ds)
        self.ds = ds
        self.transform = transform
        self.cache = multiprocessing.Manager().list([None] * len(self.ds))
        self.device = device

    def __len__(self):
        return len(self.ds)

    def __getitem__(self, i):
        """
        Returns:
            The ith element of the underlying dataset or its cached value if
            available
        """
        if self.cache[i] is None:
            self.cache[i] = tu.send_to_device(self.ds[i],
                                              self.device,
                                              non_blocking=True)

        x, *y = self.cache[i]

        if self.transform is not None:
            x = self.transform(x)

        return [x] + y

    def __repr__(self):
        return "CachedDataset:\n{}".format(tu.indent(repr(self.ds)))



from typing import Optional

import anndata
import cooler
import numpy as np
from anndata import AnnData
from scvi import settings
from scvi.data import AnnDataManager
from scvi.dataloaders import DataSplitter
from scvi.model._utils import parse_device_args
from tqdm import tqdm

from ._utils import coarsen


def _rename_chroms(c: cooler.Cooler, characters: str = "chr"):
    _rename_dict = {}
    for chromname in c.chromnames:
        if chromname.startswith(characters):
            _rename_dict[chromname] = chromname[len(characters) :]
    cooler.rename_chroms(c, _rename_dict)


def _get_contact(
    c: cooler.Cooler,
    chrom1: str,
    start1: int,
    end1: int,
    chrom2: str = None,
    start2: int = None,
    end2: int = None,
) -> np.array:
    region1 = "{}:{}-{}".format(chrom1, start1, end1)
    region2 = (
        "{}:{}-{}".format(chrom2, start2, end2) if chrom2 and start2 and end2 else None
    )

    mat = c.matrix(balance=False).fetch(region1, region2)
    return mat


def _get_cools(*cooler_uris):
    cools = [cooler.Cooler(cooler_uri) for cooler_uri in cooler_uris]
    for c in cools:
        _rename_chroms(c)

    return tuple(cools)


def setup_data(
    main_cooler_uri: str,
    sub_cooler_uri: str,
    target_cooler_uri: str,
    n_bin: int = 400,
    step_bin: int = 200,
    nonzero_ratio: float = 0.001,
) -> AnnData:
    """Setup function.

    Parameters
    ----------
    main_path
        # TODO
    sub_path
        # TODO
    target_path
        # TODO

    Returns
    -------
    An :class:`~anndata.AnnData` object containing the data required for model training.
    """
    main_c, sub_c, target_c = _get_cools(
        main_cooler_uri, sub_cooler_uri, target_cooler_uri
    )
    assert main_c.binsize == sub_c.binsize == target_c.binsize

    binsize = main_c.binsize
    chroms = set(main_c.chromnames) & set(sub_c.chromnames) & set(target_c.chromnames)

    datas = None
    for chrom in chroms:
        _datas = None
        chromsize = min(
            main_c.chromsizes[chrom],
            sub_c.chromsizes[chrom],
            target_c.chromsizes[chrom],
        )
        for i in tqdm(
            range(0, chromsize, step_bin),
            total=int(chromsize / step_bin / binsize - 1),
            desc=chrom,
        ):
            start, end = i * binsize, (i + n_bin) * binsize
            if end > chromsize:
                break

            _main = _get_contact(main_c, chrom, start, end)
            _sub = _get_contact(sub_c, chrom, start, end)
            _target = _get_contact(target_c, chrom, start, end)

            if nonzero_ratio > np.count_nonzero(_target) / np.size(_target):
                continue

            _fuzzy_main = coarsen(_main)

            data = AnnData(
                X=_fuzzy_main.flatten()[np.newaxis, :],
                layers={
                    "main(source)": _main.flatten()[np.newaxis, :],
                    "sub": _sub.flatten()[np.newaxis, :],
                    "target": _target.flatten()[np.newaxis, :],
                },
            )
            data.obs_names = ["{}:{}-{}".format(chrom, start, end)]
            _datas = anndata.concat([_datas, data]) if _datas else data
        if _datas is None:
            continue
        datas = anndata.concat([datas, _datas]) if datas else _datas

    return datas


class AnnDataSplitter(DataSplitter):
    """Creates data loaders ``train_set``, ``validation_set``, ``test_set`` using given indices."""

    def __init__(
        self,
        adata_manager: AnnDataManager,
        train_indices,
        valid_indices,
        test_indices,
        accelerator: str = "auto",
        **kwargs,
    ):
        super().__init__(adata_manager)
        self.data_loader_kwargs = kwargs
        self.accelerator = accelerator
        self.train_idx = train_indices
        self.val_idx = valid_indices
        self.test_idx = test_indices

    def setup(self, stage: Optional[str] = None):
        """Over-ride parent's setup to preserve split idx."""
        accelerator, _, self.device = parse_device_args(
            self.accelerator, return_device="torch"
        )
        self.pin_memory = (
            True
            if (settings.dl_pin_memory_gpu_training and accelerator == "gpu")
            else False
        )



import torch

from torchelie.nn import *
import torchelie.nn.functional as tnnf


def test_adain():
    m = torch.jit.script(AdaIN2d(16, 8))
    m(torch.randn(5, 16, 8, 8), torch.randn(5, 8))

    m = torch.jit.script(AdaIN2d(16, 8))
    m.condition(torch.randn(5, 8))
    m(torch.randn(5, 16, 8, 8))


def test_film():
    m = torch.jit.script(FiLM2d(16, 8))
    m(torch.randn(5, 16, 8, 8), torch.randn(5, 8))


def test_bn():
    for M in [NoAffineBN2d, NoAffineMABN2d, BatchNorm2d, MovingAverageBN2d]:
        m = torch.jit.script(M(16))
        m(torch.randn(5, 16, 8, 8))

    for M in [ConditionalBN2d, ConditionalMABN2d]:
        m = M(16, 8)
        m(torch.randn(5, 16, 8, 8), torch.randn(5, 8))

    m = torch.jit.script(PixelNorm())
    m(torch.randn(5, 16, 8, 8))

    m = Lambda(lambda x: x + 1)
    m(torch.zeros(1))


def test_spade():
    for M in [Spade2d, SpadeMA2d]:
        m = M(16, 8, 4)
        m(torch.randn(5, 16, 8, 8), torch.randn(5, 8, 8, 8))


def test_attnnorm():
    m = AttenNorm2d(16, 8)
    m(torch.randn(1, 16, 8, 8))


def test_blocks():
    m = MConvBNReLU(4, 8, 3)
    m(torch.randn(1, 4, 8, 8))

    m = ConvBlock(4, 8, 3)
    m(torch.randn(1, 4, 8, 8))

    m = ConvBlock(4, 8, 3).remove_batchnorm().leaky()
    m(torch.randn(1, 4, 8, 8))

    m = ResBlock(4, 8, 1)
    m(torch.randn(1, 4, 8, 8))

    m = PreactResBlock(4, 8, 1)
    m(torch.randn(1, 4, 8, 8))

    m = SpadeResBlock(4, 4, 3, 1)
    m(torch.randn(1, 4, 8, 8), torch.randn(1, 3, 8, 8))

    m = SpadeResBlock(4, 8, 3, 1)
    m(torch.randn(1, 4, 8, 8), torch.randn(1, 3, 8, 8))

    m = AutoGANGenBlock(6, 3, [])
    m(torch.randn(1, 6, 8, 8))

    m = AutoGANGenBlock(3, 3, [])
    m(torch.randn(1, 3, 8, 8))

    m = AutoGANGenBlock(3, 3, [5])
    m(torch.randn(1, 3, 8, 8), [torch.randn(1, 5, 4, 4)])

    m = ResidualDiscrBlock(6, 3)
    m(torch.randn(1, 6, 8, 8))


def test_vq():
    m = VQ(8, 16, mode='nearest')
    m(torch.randn(10, 8))

    m = VQ(8, 16, mode='angular')
    m(torch.randn(10, 8))

    m = VQ(8, 16, mode='nearest', init_mode='first')
    m(torch.randn(10, 8))

    m = VQ(8, 16, mode='angular', init_mode='first')
    m(torch.randn(10, 8))


def test_tfms():
    m = torch.jit.script(ImageNetInputNorm())
    m(torch.randn(1, 3, 8, 8))


def test_maskedconv():
    m = MaskedConv2d(3, 8, 3, center=True)
    m(torch.randn(1, 3, 8, 8))
    m = TopLeftConv2d(3, 8, 3, center=True)
    m(torch.randn(1, 3, 8, 8))


def test_misc():
    m = torch.jit.script(Noise(1))
    m(torch.randn(1, 3, 8, 8))

    m = Debug('test')
    m(torch.randn(1, 3, 8, 8))

    m = torch.jit.script(Reshape(16))
    m(torch.randn(1, 4, 4))


def test_laplacian():
    x = torch.randn(5, 3, 32, 32)
    tnnf.combine_laplacians(tnnf.laplacian(x))



import torch
from typing import Optional


def log_t(x: torch.Tensor, t: float) -> torch.Tensor:
    if t == 1:
        return torch.log(x)
    return (x**(1 - t) - 1) / (1 - t)


def exp_t(x: torch.Tensor, t: float) -> torch.Tensor:
    if t == 1:
        return torch.exp(x)
    return torch.clamp(1 + (1 - t) * x, min=0)**(1 / (1 - t))


def lambdas(a: torch.Tensor, t: float, n_iters: int = 3) -> torch.Tensor:
    mu = torch.max(a, dim=1, keepdim=True).values
    a_tilde = a - mu
    for i in range(n_iters):
        za = exp_t(a_tilde, t).sum(1, keepdim=True)
        a_tilde = za**(1 - t) * (a - mu)
    return -log_t(1 / za, t) + mu


def tempered_log_softmax(x: torch.Tensor,
                         t: float,
                         n_iters: int = 3) -> torch.Tensor:
    """
    Tempered log softmax. Computes log softmax along dimension 1

    Args:
        x (tensor): activations
        t (float): temperature
        n_iters (int): number of iters to converge (default: 3

    Returns:
        result of tempered log softmax
    """
    return x - lambdas(x, t, n_iters=n_iters)


def tempered_softmax(x: torch.Tensor,
                     t: float,
                     n_iters: int = 3) -> torch.Tensor:
    """
    Tempered softmax. Computes softmax along dimension 1

    Args:
        x (tensor): activations
        t (float): temperature
        n_iters (int): number of iters to converge (default: 3

    Returns:
        result of tempered softmax
    """
    return exp_t(tempered_log_softmax(x, t, n_iters), t)


def tempered_cross_entropy(x: torch.Tensor,
                           y: torch.Tensor,
                           t1: float,
                           t2: float,
                           n_iters: int = 3,
                           weight: Optional[torch.Tensor] = None,
                           reduction: str = 'mean') -> torch.Tensor:
    """
    The bi-tempered loss from https://arxiv.org/abs/1906.03361

    Args:
        x (tensor): a tensor of batched probabilities like for cross_entropy
        y (tensor): a tensor of labels
        t1 (float): temperature 1
        t2 (float): temperature 2
        weight (tensor): a tensor that associates a weight to each class
        reduction (str): how to reduce the batch of losses: 'none', 'sum', or
            'mean'

    Returns:
        the loss
    """
    sm = tempered_log_softmax(x, t2, n_iters=n_iters)
    return tempered_nll_loss(sm, y, t1, t2, weight=weight, reduction=reduction)


def tempered_nll_loss(x: torch.Tensor,
                      y: torch.Tensor,
                      t1: float,
                      t2: float,
                      weight: Optional[torch.Tensor] = None,
                      reduction: str = 'mean') -> torch.Tensor:
    """
    Compute tempered nll loss

    Args:
        x (tensor): activations of log softmax
        y (tensor): labels
        t1 (float): temperature 1
        t2 (float): temperature 2
        weight (tensor): a tensor that associates a weight to each class
        reduction (str): how to reduce the batch of losses: 'none', 'sum', or
            'mean'
    Returns:
        the loss
    """
    x = exp_t(x, t2)
    y_hat = x[torch.arange(0, x.shape[0]).long(), y]
    out = -log_t(y_hat, t1) - (1 - torch.sum(x**(2 - t1), dim=1)) / (2 - t1)
    if weight is not None:
        out = weight[y] * out

    if reduction == 'none':
        return out
    if reduction == 'mean' and weight is not None:
        return torch.sum(out / weight[y].sum())
    if reduction == 'mean' and weight is None:
        return torch.sum(out / out.shape[0])
    if reduction == 'sum':
        return out.sum()
    assert False, f'{reduction} not a valid reduction method'


class TemperedCrossEntropyLoss(torch.nn.Module):
    """
    The bi-tempered loss from https://arxiv.org/abs/1906.03361

    Args:
        t1 (float): temperature 1
        t2 (float): temperature 2
        weight (tensor): a tensor that associates a weight to each class
        reduction (str): how to reduce the batch of losses: 'none', 'sum', or
            'mean'
    """
    def __init__(self, t1, t2, weight=None, reduction='mean'):
        super(TemperedCrossEntropyLoss, self).__init__()
        self.t1 = t1
        self.t2 = t2
        self.weight = weight
        self.reduction = reduction

    def forward(self, x, y):
        """
        Forward pass

        Args:
            x (tensor): a tensor of batched probabilities like for
                cross_entropy
            y (tensor): a tensor of labels

        Returns:
            the loss
        """
        return tempered_cross_entropy(x,
                                      y,
                                      self.t1,
                                      self.t2,
                                      weight=self.weight,
                                      reduction=self.reduction)



import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchelie.utils as tu
import torchelie as tch
from typing import Optional
from torch.autograd import Function


class AdaptiveConcatPool2d(nn.Module):
    """
    Pools with AdaptiveMaxPool2d AND AdaptiveAvgPool2d and concatenates both
    results.

    Args:
        target_size: the target output size (single integer or
            double-integer tuple)
    """
    def __init__(self, target_size):
        super(AdaptiveConcatPool2d, self).__init__()
        self.target_size = target_size

    def forward(self, x):
        return torch.cat([
            nn.functional.adaptive_avg_pool2d(x, self.target_size),
            nn.functional.adaptive_max_pool2d(x, self.target_size),
        ], dim=1)


class ModulatedConv(nn.Conv2d):
    def __init__(self,
                 in_channels: int,
                 noise_channels: int,
                 *args,
                 demodulate: bool = True,
                 **kwargs):
        super(ModulatedConv, self).__init__(in_channels, *args, **kwargs)
        self.make_s = tu.xavier(nn.Linear(noise_channels, in_channels))
        self.make_s.bias.data.fill_(1)
        self.demodulate = demodulate

    def condition(self, z: torch.Tensor) -> None:
        self.s = self.make_s(z)

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        if z is not None:
            self.condition(z)
        N, C, H, W = x.shape
        C_out, C_in = self.weight.shape[:2]
        w_prime = torch.einsum('oihw,bi->boihw', self.weight, self.s)
        if self.demodulate:
            w_prime_prime = torch.einsum('boihw,boihw->bo', w_prime, w_prime)
            w_prime_prime = w_prime_prime.add_(1e-8).rsqrt()
            w = w_prime * w_prime_prime[..., None, None, None]
        else:
            w = w_prime

        w = w.view(-1, *w.shape[2:])
        x = F.conv2d(x.view(1, -1, H, W), w, None, self.stride, self.padding,
                     self.dilation, N)
        x = x.view(N, C_out, H, W)
        if self.bias is not None:
            return x.add_(self.bias.view(-1, 1, 1))
        else:
            return x


class SelfAttention2d(nn.Module):
    """
    Self Attention such as used in SAGAN or BigGAN.

    Args:
        ch (int): number of input / output channels
    """
    def __init__(self, ch: int):
        super().__init__()
        self.key = nn.Conv1d(ch, ch // 8, 1)
        self.query = nn.Conv1d(ch, ch // 8, 1)
        self.value = nn.Conv1d(ch, ch, 1)
        self.gamma = nn.Parameter(torch.tensor([0.]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        forward
        """
        x_flat = x.view(*x.shape[:2], -1)
        k = self.key(x_flat)
        q = self.query(x_flat)
        v = self.value(x_flat)

        affinity = torch.einsum('bki,bkj->bij', q, k)
        attention = F.softmax(affinity, dim=1)
        out = torch.einsum('bci,bih->bch', v, attention).view(*x.shape)
        return self.gamma * out + x


class GaussianPriorFunc(Function):
    @staticmethod
    def forward(ctx, mu, sigma, mu2, sigma2, strength=1):
        z = torch.randn_like(mu)
        x = mu + z * sigma
        s = torch.tensor(strength, device=x.device)
        ctx.save_for_backward(mu, sigma, z, mu2, sigma2, s)
        return x

    @staticmethod
    def backward(ctx, d_out):
        mu, sigma, z, mu2, sigma2, strength = ctx.saved_tensors

        # kl = -0.5
        #      + log(sigma2) - log(sigma1)
        #      + (sigma1 ** 2) / (2 * sigma2 ** 2)
        #      + ((mu1 - mu2) ** 2) / (2 * sigma2 ** 2)
        # dkl/dmu  =  (mu - mu2) / (sigma**2)
        # dkl/dmu2 = -(mu - mu2) / (sigma**2)
        # dkl/dsig  = -1/sigma + sigma/(sigma2**2)
        # dkl/dsig2 = 1/sigma2 - ((sigma**2) + ((mu-mu2)**2)) / (sigma2 ** 3)
        diff_mu = mu - mu2
        s2sq = sigma2.pow(2)
        diff_mu_over_s2sq = diff_mu / s2sq
        d_mu = d_out + strength * diff_mu_over_s2sq
        d_mu2 = -strength * diff_mu_over_s2sq
        d_sigma = d_out * z - strength / sigma + strength * sigma / s2sq
        d_sigma2 = 1 / sigma2 - (sigma.pow(2) + diff_mu.pow(2)) / sigma2.pow(3)
        d_sigma2 *= strength
        return d_mu, d_sigma, d_mu2, d_sigma2


class UnitGaussianPrior(nn.Module):
    """
    Force a representation to fit a unit gaussian prior. It projects with a
    nn.Linear the input vector to a mu and sigma that represent a gaussian
    distribution from which the output is sampled. The backward pass includes a
    kl divergence loss between N(mu, sigma) and N(0, 1).

    This can be used to implement VAEs or information bottlenecks

    In train mode, the output is sampled from N(mu, sigma) but at test time mu
    is returned.

    Args:
        in_channels (int): dimension of input channels
        num_latents (int): dimension of output latents
        strength (float): strength of the kl loss. When using this to implement
            a VAE, set strength to :code:`1/number of output dim of the model`
            or set it to 1 but make sure that the loss for each output
            dimension is summed, but averaged over the batch.
        kl_reduction (str): how the implicit kl loss is reduced over the batch
            samples. 'sum' means the kl term of each sample is summed, while
            'mean' divides the loss by the number of examples.
    """
    def __init__(self,
                 in_channels,
                 num_latents,
                 strength=1,
                 kl_reduction='mean'):
        super().__init__()
        self.project = tu.kaiming(nn.Linear(in_channels, 2 * num_latents))
        self.project.bias.data[num_latents:].fill_(1)
        self.strength = strength
        assert kl_reduction in ['mean', 'sum']
        self.reduction = kl_reduction

    def forward(self, x):
        """
        Args:
            x (Tensor): A 2D (N, in_channels) tensor

        Returns:
            A 2D (N, num_channels) tensor sampled from the implicit gaussian
                distribution.
        """
        x = self.project(x)
        mu, sigma = torch.chunk(x, 2, dim=1)
        if self.training:
            sigma = torch.exp(0.5 * sigma)
            strength = self.strength
            if self.reduction == 'mean':
                strength = strength / x.shape[0]
            return tch.nn.functional.unit_gaussian_prior(mu, sigma, strength)
        else:
            return mu


class InformationBottleneck(UnitGaussianPrior):
    pass


@tu.experimental
class Const(nn.Module):
    """
    Return a constant learnable volume. Disregards the input except its batch
    size

    Args:
        *size (ints): the shape of the volume to learn
    """
    def __init__(self, *size: int) -> None:
        super().__init__()
        self.size = size
        self.weight = nn.Parameter(torch.randn(1, *size))

    def extra_repr(self):
        return repr(self.size)

    def forward(self, n: int) -> torch.Tensor:
        """
        Args:
            n (int): batch size to use
        """
        return self.weight.expand(n, *self.weight.shape[1:]).contiguous()


@tu.experimental
class SinePositionEncoding2d(nn.Module):
    def __init__(self, n_fourier_freqs: int) -> None:
        super().__init__()
        self.register_buffer('fourier_freqs', torch.randn(n_fourier_freqs, 2, 1, 1))

    def forward(self, x):
        h = torch.arange(0, x.shape[2] * 0.1, 0.1)
        v = torch.arange(0, x.shape[3] * 0.1, 0.1)
        hv = torch.stack(torch.meshgrid(h, v), dim=0)[None]
        out = F.conv2d(hv.to(x.device, x.dtype), self.fourier_freqs)
        out = torch.cat([torch.sin(out), torch.cos(out)], dim=1)
        out /= math.sqrt(out.shape[1])
        out = torch.cat([x, out.expand(x.shape[0], -1, -1, -1)], dim=1)
        return out


class MinibatchStddev(nn.Module):
    """Minibatch Stddev layer from Progressive GAN"""
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        stddev_map = torch.sqrt(x.var(dim=0) + 1e-8).mean()
        stddev = stddev_map.expand(x.shape[0], 1, *x.shape[2:])
        return torch.cat([x, stddev], dim=1)


class HardSigmoid(nn.Module):
    """
    Hard Sigmoid
    """
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.add_(0.5).clamp_(min=0, max=1)


class HardSwish(nn.Module):
    """
    Hard Swish
    """
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.add(0.5).clamp_(min=0, max=1).mul_(x)



r"""
Standard, non saturating, GAN loss from the original GAN paper

https://arxiv.org/abs/1406.2661

:math:`L_D(x_r, x_f) = - \log(1 - D(x_f)) - \log D(x_r)`

:math:`L_G(x_f) = -\log D(x_f)`
"""

import torch
import torch.nn.functional as F


def real(x: torch.Tensor, reduce: str = 'mean') -> torch.Tensor:
    out = F.softplus(-x)
    if reduce == 'none':
        return out
    if reduce == 'mean':
        return out.mean()
    if reduce == 'batch_mean':
        return out.view(out.shape[0], -1).sum(1).mean()
    assert False, f'reduction {reduce} invalid'


def fake(x: torch.Tensor, reduce: str = 'mean') -> torch.Tensor:
    out = F.softplus(x)
    if reduce == 'none':
        return out
    if reduce == 'mean':
        return out.mean()
    if reduce == 'batch_mean':
        return out.view(out.shape[0], -1).sum(1).mean()
    assert False, f'reduction {reduce} invalid'


def generated(x: torch.Tensor, reduce: str = 'mean') -> torch.Tensor:
    out = F.softplus(-x)
    if reduce == 'none':
        return out
    if reduce == 'mean':
        return out.mean()
    if reduce == 'batch_mean':
        return out.view(out.shape[0], -1).sum(1).mean()
    assert False, f'reduction {reduce} invalid'



import torch
import torch.nn as nn
from torch.distributions import TransformedDistribution


class Logistic(TransformedDistribution):
    """
    Logistic distribution

    Args:
        loc (tensor): mean of the distribution
        scale (tensor): scale of the distribution
    """
    def __init__(self, loc: torch.Tensor, scale: torch.Tensor) -> None:
        td = torch.distributions
        base_distribution = td.Uniform(torch.zeros_like(loc),
                                       torch.ones_like(loc))
        transforms = [
            td.SigmoidTransform().inv,
            td.AffineTransform(loc=loc, scale=scale)
        ]
        super(Logistic, self).__init__(base_distribution, transforms)


class LogisticMixture:
    """
    Mixture of Logistic distributions. Each tensor contains an additional
    dimension with `number of distributions` elements.

    Args:
        weights (tensor): un-normalized weights of distributions
        loc (tensor): mean of the distributions
        scale (tensor): scale of the distributions
        dim (int): dimension reprenseting the various distributions, that will
            weighted and averaged on.
    """
    def __init__(self, weights, locs, scales, dim) -> None:
        self.weights = weights
        self.logistics = Logistic(locs, scales)
        self.locs = locs
        self.dim = dim - len(locs.shape) if dim >= 0 else dim

    @property
    def mean(self) -> torch.Tensor:
        w = nn.functional.softmax(self.weights, dim=self.dim)
        return torch.sum(self.locs * w, dim=self.dim)

    def log_prob(self, x: torch.Tensor) -> torch.Tensor:
        log_pis = nn.functional.log_softmax(self.weights, dim=self.dim)
        return torch.logsumexp(self.logistics.log_prob(x.unsqueeze(self.dim))
                               + log_pis,
                               dim=self.dim)


class GaussianMixture:
    """
    Mixture of gaussian distributions. Each tensor contains an additional
    dimension with `number of distributions` elements.

    Args:
        weights (tensor): un-normalized weights of distributions
        loc (tensor): mean of the distributions
        scale (tensor): scale of the distributions
        dim (int): dimension reprenseting the various distributions, that will
            weighted and averaged on.
    """
    def __init__(self, weights: torch.Tensor, locs: torch.Tensor,
                 scales: torch.Tensor) -> None:
        self.weights = weights
        self.logistics = torch.distributions.Normal(locs, scales)
        self.locs = locs

    @property
    def mean(self) -> torch.Tensor:
        w = nn.functional.softmax(self.weights, dim=1)
        return torch.sum(self.locs * w, dim=1)

    def log_prob(self, x: torch.Tensor) -> torch.Tensor:
        log_pis = nn.functional.log_softmax(self.weights, dim=1)
        return torch.logsumexp(self.logistics.log_prob(x.unsqueeze(1))
                               + log_pis,
                               dim=1)



from typing import Tuple
import torch


def zero_gp(model,
            real: torch.Tensor,
            fake: torch.Tensor,
            amp_scaler=None) -> Tuple[torch.Tensor, float]:
    """
    0-GP from Improving Generalization And Stability Of Generative Adversarial
    Networks ( https://arxiv.org/abs/1902.03984 ).

    Args:
        model (function / nn.Module): the model
        real (torch.Tensor): real images
        fake (torch.Tensor): fake images
        amp_scaler (torch.cuda.amp.GradScaler): if specified, will be used
            for computing the loss in fp16. Otherwise, use model's and
            data's dtype.

    Returns:
        A tuple (loss, gradient norm)
    """
    t = torch.rand(real.shape[0], 1, 1, 1, device=real.device, dtype=real.dtype)
    x = t * real + (1 - t) * fake
    return gradient_penalty(model, x, 0., amp_scaler)


def wgan_gp(model,
            real: torch.Tensor,
            fake: torch.Tensor,
            amp_scaler=None) -> Tuple[torch.Tensor, float]:
    """
    1-GP from Improved Training of Wasserstein GANs (
    https://arxiv.org/abs/1704.00028 ).

    Args:
        model (function / nn.Module): the model
        real (torch.Tensor): real images
        fake (torch.Tensor): fake images
        amp_scaler (torch.cuda.amp.GradScaler): if specified, will be used
            for computing the loss in fp16. Otherwise, use model's and
            data's dtype.

    Returns:
        A tuple (loss, gradient norm)
    """
    t = torch.rand(real.shape[0], 1, 1, 1, device=real.device)
    t = torch.rand(real.shape[0], 1, 1, 1, device=real.device)
    x = t * real + (1 - t) * fake
    return gradient_penalty(model, x, 1., amp_scaler)


def R1(model,
       real: torch.Tensor,
       fake: torch.Tensor,
       amp_scaler=None) -> Tuple[torch.Tensor, float]:
    """
    R1 regularizer from Which Training Methods for GANs do actually Converge?
    ( https://arxiv.org/abs/1801.04406 ).

    Args:
        model (function / nn.Module): the model
        real (torch.Tensor): real images
        fake (torch.Tensor): unused. Here for interface consistency with other
            penalties.
        amp_scaler (torch.cuda.amp.GradScaler): if specified, will be used
            for computing the loss in fp16. Otherwise, use model's and
            data's dtype.

    Returns:
        A tuple (loss, gradient norm)
    """
    return gradient_penalty(model, real, 0., amp_scaler)


def gradient_penalty(model,
                     data: torch.Tensor,
                     objective_norm: float,
                     amp_scaler=None) -> Tuple[torch.Tensor, float]:
    """
    Gradient penalty, mainly for GANs. Of the form
    :code:`E[(||dmodel(data)/ddata|| - objective_goal)²]`

    Args:
        model (function / nn.Module): the model
        data (torch.Tensor): input on which to measure the norm of the
            gradient
        objective_norm (float): objective norm. 1 for WGAN GP for instance.
        amp_scaler (torch.cuda.amp.GradScaler): if specified, will be used
            for computing the loss in fp16. Otherwise, use model's and
            data's dtype.

    Returns:
        A tuple (loss, gradient norm)
    """
    data = data.detach()
    data.requires_grad_(True)

    fp16 = amp_scaler is not None and amp_scaler.is_enabled()
    with torch.cuda.amp.autocast(enabled=fp16):
        out = model(data).sum()

    scale = amp_scaler.scale if fp16 else (lambda x: x)
    g = torch.autograd.grad(outputs=scale(out),
                            inputs=data,
                            create_graph=True,
                            only_inputs=True)[0]

    g = (g / amp_scaler.get_scale()) if fp16 else g

    g_norm = g.pow(2).sum(dim=(1, 2, 3)).add_(1e-8).sqrt()
    return (g_norm - objective_norm).pow(2).mean(), g_norm.mean().item()
